from xoa_utils.clis.cli_utils import (
    ReadConfig,
    run_coroutine_as_sync,
    format_tester_status,
    format_ports_status,
    format_an_status,
    format_port_status,
    format_recovery,
    format_lt_algorithm,
    format_lt_config,
    format_lt_im,
    format_error,
    format_an_config,
    format_lt_inc_dec,
    format_lt_encoding,
    format_lt_preset,
    format_lt_trained,
    format_txtap_get,
    format_txtap_set,
    format_lt_status,
    format_debug_init,
    format_strict,
    format_log_control,
)



from __future__ import annotations
import asyncclick as ac
import asyncio
from xoa_utils.clicks import click_backend as cb
from xoa_driver.hlfuncs import anlt as anlt_utils
from xoa_driver.hlfuncs import mgmt as mgmt_utils
from xoa_driver.testers import L23Tester
from xoa_driver.enums import MediaConfigurationType
from xoa_utils.clis import format_tester_status, format_ports_status, format_port_status
from xoa_utils.clicks.click_commands.group import xoa_util
from xoa_utils.clicks import click_help as h
from xoa_utils.cmds import CmdContext
from xoa_utils import exceptions as ex

# --------------------------
# command: connect
# --------------------------
@xoa_util.command(cls=cb.XenaCommand)
@ac.argument("device", type=ac.STRING)
@ac.argument("username", type=ac.STRING)
@ac.option("-p", "--ports", type=ac.STRING, help=h.HELP_CONNECT_PORT_LIST, default="")
@ac.option("--reset/--no-reset", type=ac.BOOL, help=h.HELP_CONNECT_RESET, default=True)
@ac.option("--force/--no-force", type=ac.BOOL, help=h.HELP_CONNECT_FORCE, default=True)
@ac.option("-P", "--password", type=ac.STRING, help=h.HELP_CONNECT_PWD, default="xena")
@ac.option("-t", "--tcp", type=ac.INT, help=h.HELP_CONNECT_TCP_PORT, default=22606)
@ac.pass_context
async def connect(
    context: ac.Context,
    device: str,
    username: str,
    ports: str,
    reset: bool,
    force: bool,
    password: str,
    tcp: int,
) -> str:
    """
    Connect to a tester for the current session.

        <DEVICE>: The chassis address to connect. Address can be in IPv4 format (e.g. 10.10.10.10), or a host name (e.g. demo.xenanetworks.com)

        <USERNAME>: Username for port reservation.
    """
    storage: CmdContext = context.obj
    real_port_list = [i.strip() for i in ports.split(",")] if ports else []
    tester = await L23Tester(device, username, password, tcp, debug=False)
    con_info = f"{device}:{tcp}"
    storage.store_current_tester(username, con_info, tester)
    count = 0
    first_id = ""
    for id_str in real_port_list:
        this_module_dic = storage.obtain_physical_modules(id_str)
        for module_id, module_obj in this_module_dic.items():
            storage.store_module(module_id, module_obj)
        this_port_dic = storage.obtain_physical_ports(id_str)
        for port_id, port_obj in this_port_dic.items():
            if force:
                await mgmt_utils.reserve_port(port_obj, force)
            if reset:
                await mgmt_utils.reset_port(port_obj)
            port_serdes_num = (await anlt_utils.anlt_status(port_obj))["serdes_count"]
            storage.store_port(port_id, port_obj, port_serdes_num)
            if count == 0:
                first_id = port_id
            count += 1
    if real_port_list:
        storage.store_current_module_str(first_id.split("/")[0])
        storage.store_current_port_str(first_id)
    if force or reset:
        await asyncio.sleep(2)
        # status will change when you reserve_port or reset_port, need to wait
    return format_tester_status(storage)


# --------------------------
# command: exit
# --------------------------
@xoa_util.command(cls=cb.XenaCommand)
@ac.option("--reset/--no-reset", type=ac.BOOL, help=h.HELP_CONNECT_RESET, default=True)
@ac.option(
    "--release/--no-release", type=ac.BOOL, help=h.HELP_EXIT_RELEASE, default=True
)
@ac.pass_context
async def exit(context: ac.Context, reset: bool, release: bool) -> str:
    """
    Exit the session. Exit by terminating port reservations, disconnecting from the chassis, releasing system resources, and removing the specified port configurations.
    """
    storage: CmdContext = context.obj
    for module_id, module_obj in storage.retrieve_modules().copy().items():
        if release:
            await mgmt_utils.free_module(module_obj)
        storage.remove_module(module_id)
    for port_id, port_obj in storage.retrieve_ports().copy().items():
        if reset:
            await mgmt_utils.reset_port(port_obj)
        if release:
            await mgmt_utils.free_port(port_obj)
        storage.remove_port(port_id)

    return ""


# --------------------------
# command: port
# --------------------------
@xoa_util.command(cls=cb.XenaCommand)
@ac.argument("port", type=ac.STRING)
@ac.option("--reset/--no-reset", type=ac.BOOL, help=h.HELP_CONNECT_RESET, default=False)
@ac.option("--force/--no-force", type=ac.BOOL, help=h.HELP_CONNECT_FORCE, default=True)
@ac.pass_context
async def port(context: ac.Context, port: str, reset: bool, force: bool) -> str:
    """
    Switch the working port. If the port is not yet reserved, reserve the port. Update the working port in the cache.

        <PORT>: The port on the specified device host. Specify a port using the format slot/port, e.g. 0/0
    """
    storage: CmdContext = context.obj
    module_id = port.split("/")[0]
    try:
        storage.store_current_module_str(module_id)
        p_obj = storage.retrieve_module(module_id)
    except ex.NotInStoreError:
        module_dic = storage.obtain_physical_modules(module_id)
        for m_id, m_obj in module_dic.items():
            storage.store_module(m_id, m_obj)
            storage.store_current_module_str(m_id)
    try:
        storage.store_current_port_str(port)
        p_obj = storage.retrieve_port()
        port_serdes_num = (await anlt_utils.anlt_status(p_obj))["serdes_count"]
        storage.store_port(port, p_obj, port_serdes_num)
    except ex.NotInStoreError:
        port_dic = storage.obtain_physical_ports(port)
        for p_id, p_obj in port_dic.items():
            port_serdes_num = (await anlt_utils.anlt_status(p_obj))["serdes_count"]
            storage.store_port(p_id, p_obj, port_serdes_num)
            storage.store_current_port_str(p_id)
    port_obj = storage.retrieve_port()
    port_id = storage.retrieve_port_str()
    tester_obj = storage.retrieve_tester()

    if force:
        module_obj = mgmt_utils.get_module(tester_obj, int(module_id))
        await mgmt_utils.free_module(module_obj)
        await mgmt_utils.reserve_port(port_obj, force)
    if reset:
        await mgmt_utils.reset_port(port_obj)
    if force or reset:
        await asyncio.sleep(2)
        # status will change when you reserve_port or reset_port, need to wait
    status_dic = await anlt_utils.anlt_status(port_obj)
    return f"{format_ports_status(storage, False)}{format_port_status(status_dic, storage)}"


# --------------------------
# command: ports
# --------------------------
@xoa_util.command(cls=cb.XenaCommand)
@ac.option("--all/--no-all", type=ac.BOOL, help=h.HELP_PORTS_ALL, default=False)
@ac.pass_context
async def ports(context: ac.Context, all: bool) -> str:
    """
    List all the ports reserved by the current session.
    """
    storage: CmdContext = context.obj
    return format_ports_status(storage, all)


# --------------------------
# command: module-config
# --------------------------
@xoa_util.command(cls=cb.XenaCommand)
@ac.argument("module", type=ac.INT)
@ac.argument(
    "media",
    type=ac.Choice(
        [
            "cfp4",
            "cxp",
            "sfp28",
            "qsfp28_nrz",
            "qsfp28_pam4",
            "qsfp56_pam4",
            "qsfpdd_pam4",
            "sfp56",
            "sfpdd",
            "sfp112",
            "qsfpdd_nrz",
            "cfp",
            "base_t1",
            "base_t1s",
            "qsfpdd800",
            "qsfp112",
            "osfp800",
        ]
    ),
)
@ac.argument("port_count", type=ac.INT)
@ac.argument(
    "port_speed",
    type=ac.Choice(
        [
            "800g",
            "400g",
            "200g",
            "100g",
        ]
    ),
)
@ac.option("--force/--no-force", type=ac.BOOL, help=h.HELP_CONNECT_FORCE, default=True)
@ac.pass_context
async def module_config(
    context: ac.Context,
    module: int,
    media: str,
    port_count: int,
    port_speed: str,
    force: bool,
) -> str:
    """
    Change module media configuration

        <MODULE>: Specifies the module on the specified device host. Specify a module using the format slot, e.g. 0

        <MEDIA>: Specifies the media configuration type of the module. Allowed values: qsfpddpam4 | sfpdd | sfp112 | qsfpddnrz | qsfpdd800 | qsfp112 | osfp800

        <PORT_COUNT>: Specifies the port count of the module.

        <PORT_SPEED>: Specifies the port speed in Gbps of the module. Allowed values: 800g | 400g | 200g | 100g

    """
    storage: CmdContext = context.obj
    module_obj = storage.retrieve_module(str(module))
    await mgmt_utils.set_module_media_config(
        module_obj, MediaConfigurationType[media.upper()], force
    )
    await mgmt_utils.set_module_port_config(
        module_obj, port_count, int(port_speed.replace("g", "000000000")), force
    )
    await module_obj._setup()
    storage.remove_ports()
    return ""



from __future__ import annotations
import asyncclick as ac
import json
import typing as t
from xoa_driver.hlfuncs import anlt as anlt_utils
from xoa_utils.clicks import click_backend as cb
from xoa_utils.clis import (
    format_recovery,
    format_port_status,
    format_strict,
    format_log_control
    )
from xoa_driver.enums import AnLtLogControl
from xoa_utils.clicks.click_commands.group import xoa_util
from xoa_utils.clicks import click_help as h
from xoa_utils.cmds import CmdContext
from enum import Enum


class ASCIIStyle(Enum):
    DARKRED = "\033[31m"
    DARKGREEN = "\033[32m"
    DARKYELLOW = "\033[33m"
    DARKBLUE = "\033[34m"
    DARKMAGENTA = "\033[35m"
    DARKCYAN = "\033[36m"

    RED = "\033[91m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    BLUE = "\033[94m"
    MAGENTA = "\033[95m"
    CYAN = "\033[96m"

    DARKRED_BG = "\033[41m"
    DARKGREEN_BG = "\033[42m"
    DARKYELLOW_BG = "\033[43m"
    DARKBLUE_BG = "\033[44m"
    DARKMAGENTA_BG = "\033[45m"
    DARKCYAN_BG = "\033[46m"

    RED_BG = "\033[101m"
    GREEN_BG = "\033[102m"
    YELLOW_BG = "\033[103m"
    BLUE_BG = "\033[104m"
    MAGENTA_BG = "\033[105m"
    CYAN_BG = "\033[106m"

    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


@xoa_util.group(cls=cb.XenaGroup)
def anlt():
    """
    Commands for AN/LT.
    """


# --------------------------
# command: recovery
# --------------------------
@anlt.command(cls=cb.XenaCommand)
@ac.option("--on/--off", type=ac.BOOL, help=h.HELP_RECOVERY_ON, default=True)
@ac.pass_context
async def recovery(context: ac.Context, on: bool) -> str:
    """
    Enable/disable link recovery on the working port.
    If enable, the port will keep doing AN/LT when no link-up signal is detected.
    """
    storage: CmdContext = context.obj

    port_obj = storage.retrieve_port()
    await anlt_utils.anlt_link_recovery(port_obj, on)
    return format_recovery(storage, on)


# --------------------------
# command: status
# --------------------------
@anlt.command(cls=cb.XenaCommand)
@ac.pass_context
async def status(context: ac.Context) -> str:
    """
    Show AN/LT status of the working port.
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    status_dic = await anlt_utils.anlt_status(port_obj)
    port_id = storage.retrieve_port_str()
    return format_port_status(status_dic, storage)


# --------------------------
# command: do
# --------------------------
@anlt.command(cls=cb.XenaCommand)
@ac.pass_context
async def do(context: ac.Context) -> str:
    """
    Apply and start AN/LT to the working port.
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    an_enable = storage.retrieve_an_enable()
    lt_enable = storage.retrieve_lt_enable()
    an_allow_loopback = storage.retrieve_an_loopback()
    lt_preset0 = storage.retrieve_lt_preset0()
    lt_initial_modulations = storage.retrieve_lt_initial_mod()
    lt_interactive = storage.retrieve_lt_interactive()
    lt_algorithm = storage.retrieve_lt_algorithm()
    await anlt_utils.anlt_start(
        port_obj,
        an_enable,
        lt_enable,
        an_allow_loopback,
        lt_preset0,
        lt_initial_modulations,
        lt_interactive,
        lt_algorithm,
    )
    return ""


# --------------------------
# command: log
# --------------------------
@anlt.command(cls=cb.XenaCommand, name="log")
@ac.option(
    "-f", "--filename", type=ac.STRING, help=h.HELP_ANLT_LOG_FILENAME, default=""
)
@ac.option(
    "-k",
    "--keep",
    type=ac.Choice(["all", "an", "lt"]),
    help=h.HELP_ANLT_LOG_KEEP,
    default="all",
)
@ac.option("-s", "--serdes", type=ac.STRING, help=h.HELP_ANLT_LOG_SERDES, default="")
@ac.pass_context
async def anlt_log(ctx: ac.Context, filename: str, keep: str, serdes: str) -> str:
    """
    Start AN/LT logging.
    """

    def _filter_log(log: str, keep: str, serdes: list[int]) -> list[dict]:
        all_logs = []
        for lg in log.split("\n"):
            try:
                content = json.loads(lg)
                log_serdes = content["lane"]
                module = content["module"]

                serdes_in = (serdes and log_serdes in serdes) or (not serdes)
                keep_in = any(
                    (
                        keep == "an" and "AN" in module,
                        keep == "lt" and "LT" in module,
                        keep == "all",
                    )
                )
                if serdes_in and keep_in:
                    all_logs.append(content)

            except Exception:
                pass
        return all_logs

    def _dict_get(dic: dict, *keys: str) -> t.Any:
        current = dic
        for k in keys:
            current = current.get(k, "")
            if current == "":
                break
        return current

    def _flatten(dic: dict[str, str]) -> str:
        return "".join((f"{k}: {v:<7}" for k, v in dic.items()))

    def _ascii_styler(str: str, fg_style: list[ASCIIStyle]) -> str:
        style = "".join(s.value for s in fg_style)
        return f"{style}{str}{ASCIIStyle.END.value}"

    def _beautify(filtered: list[dict]) -> str:
        real = []
        for i in filtered:
            b_str = ""

            log_time = _dict_get(i, "time")
            log_entry = _dict_get(i, "entry")
            log_type = _dict_get(i, "type")
            log_m = _dict_get(i, "module")
            log_log = _dict_get(i, "entry", "log")
            log_serdes = _dict_get(i, "lane")
            log_m = _dict_get(i, "module")
            log_m = _dict_get(i, "module")
            log_event = _dict_get(i, "entry", "fsm", "event")
            log_current = _dict_get(i, "entry", "fsm", "current")
            log_new = _dict_get(i, "entry", "fsm", "new")
            log_direction = _dict_get(i, "entry", "direction")
            if log_direction == "tx":
                log_direction = _ascii_styler(
                    log_direction.upper(), [ASCIIStyle.DARKBLUE_BG]
                )
            else:
                log_direction = _ascii_styler(
                    log_direction.upper(), [ASCIIStyle.DARKGREEN_BG]
                )

            log_value = _dict_get(i, "entry", "pkt", "value")
            log_ptype = _dict_get(i, "entry", "pkt", "type")
            log_pstate = _dict_get(i, "entry", "pkt", "state")
            log_np = _dict_get(i, "entry", "pkt", "fields", "NP")
            log_ack = _dict_get(i, "entry", "pkt", "fields", "Ack")
            log_rf = _dict_get(i, "entry", "pkt", "fields", "RF")
            log_tn = _dict_get(i, "entry", "pkt", "fields", "TN")
            log_en = _dict_get(i, "entry", "pkt", "fields", "EN")
            log_c = _dict_get(i, "entry", "pkt", "fields", "C")
            log_fec = _dict_get(i, "entry", "pkt", "fields", "fec")
            log_ab = _dict_get(i, "entry", "pkt", "fields", "ability")
            log_pkt_ctrl = _dict_get(i, "entry", "pkt", "fields", "control")
            log_pkt_status = _dict_get(i, "entry", "pkt", "fields", "status")
            log_pkt_locked = _dict_get(i, "entry", "pkt", "fields", "locked")
            log_mp = _dict_get(i, "entry", "pkt", "fields", "MP")
            log_ack2 = _dict_get(i, "entry", "pkt", "fields", "Ack2")
            log_t = _dict_get(i, "entry", "pkt", "fields", "T")
            log_fmt_v = _dict_get(
                i, "entry", "pkt", "fields", "formatted message", "value"
            )
            log_fmt_msg = _dict_get(
                i, "entry", "pkt", "fields", "formatted message", "message"
            )
            log_ufmt_v = _dict_get(
                i, "entry", "pkt", "fields", "un-formatted message", "value"
            )
            log_ufmt_msg = _dict_get(
                i, "entry", "pkt", "fields", "un-formatted message", "message"
            )
            log_ufmt_fec = _dict_get(
                i, "entry", "pkt", "fields", "un-formatted message", "fec"
            )
            log_ufmt_ab = _dict_get(
                i, "entry", "pkt", "fields", "un-formatted message", "ability"
            )
            log_errors = _dict_get(i, "entry", "pkt", "errors")

            if log_pkt_locked == "true":
                log_pkt_locked = _ascii_styler(log_pkt_locked, [ASCIIStyle.GREEN_BG])
            else:
                log_pkt_locked = _ascii_styler(log_pkt_locked, [ASCIIStyle.RED_BG])

            log_pkt_done = _dict_get(i, "entry", "pkt", "fields", "done")
            if log_pkt_done == "true":
                log_pkt_done = _ascii_styler(log_pkt_done, [ASCIIStyle.GREEN_BG])
            else:
                log_pkt_done = _ascii_styler(log_pkt_done, [ASCIIStyle.RED_BG])

            log_pkt_value = _dict_get(i, "entry", "pkt", "value")

            serdes_str = f"(S{log_serdes})," if "LT" in log_m else ","
            common = f"{log_time/1000000:.6f}, {log_m}{serdes_str}"

            if log_type == "debug":
                b_str = f"{common:<32}{'DBG:':<5}{log_log}"
            elif log_type == "fsm":
                b_str = (
                    f"{common:<32}{'FSM:':<5}({log_event}) {log_current} -> {log_new}"
                )
            elif log_type == "trace" and "log" in log_entry:
                b_str = f"{common:<32}{'MSG:':<5}{log_log}"
            elif log_type == "trace" and "direction" in log_entry and "LT" not in log_m:
                if log_pstate == "new" or log_pstate == "":
                    if log_ptype == "base page":
                        b_str = f"{common:<32}{(log_direction + ':'):<14}{log_value}, {log_ptype}, NP:{int(log_np, 0)}, ACK:{int(log_ack, 0)}, RF:{int(log_rf, 0)}, TN:{int(log_tn, 0)}, EN:{int(log_en ,0)}, C:{int(log_c, 0)}\n{'':<37}FEC:{log_fec}, ABILITY:{log_ab}"
                    else:
                        if log_fmt_v:
                            b_str = f"{common:<32}{(log_direction + ':'):<14}{log_value}, {log_ptype}, NP:{int(log_np, 0)}, ACK:{int(log_ack, 0)}, MP:{int(log_mp, 0)}, ACK2:{int(log_ack2, 0)}, T:{int(log_t ,0)}\n{'':<37}Formatted message:\n{'':<37}Value:{log_fmt_v}, Msg:{log_fmt_msg}"
                        else:
                            b_str = f"{common:<32}{(log_direction + ':'):<14}{log_value}, {log_ptype}, NP:{int(log_np, 0)}, ACK:{int(log_ack, 0)}, MP:{int(log_mp, 0)}, ACK2:{int(log_ack2, 0)}, T:{int(log_t ,0)}\n{'':<37}Un-formatted message:\n{'':<37}Value:{log_ufmt_v}, Msg:{log_ufmt_msg}\n{'':<37}FEC:{log_ufmt_fec}, ABILITY:{log_ufmt_ab}"
                    if log_errors:
                        b_str += "\n" + f"{'':<37}" + _ascii_styler("ERRORS:", [ASCIIStyle.RED_BG]) + f"{log_errors}"

            elif log_type == "trace" and "direction" in log_entry and "LT" in log_m:
                if log_pstate == "new" or log_pstate == "":
                    b_str = f"{common:<32}{(log_direction + ':'):<14}{log_pkt_value}, LOCKED={log_pkt_locked}, TRAINED={log_pkt_done}\n{'':<37}{_flatten(log_pkt_ctrl)}\n{'':<37}{_flatten(log_pkt_status)}"
                    if log_errors:
                        b_str += "\n" + f"{'':<37}" + _ascii_styler("ERRORS:", [ASCIIStyle.RED_BG]) + f"{log_errors}"
            elif log_type == "xla":
                log_xla = _ascii_styler("XLA", [ASCIIStyle.RED_BG])
                b_str = f"{common:<32}{(log_xla + ': '):<5}{log_log}"

            if b_str:
                real.append(b_str)
        return "\n".join(real)

    async def log(
        storage: CmdContext, filename: str, keep: str, serdes: list[int]
    ) -> str:
        port_obj = storage.retrieve_port()
        log_str = await anlt_utils.anlt_log(port_obj)
        filtered = _filter_log(log_str, keep, serdes)
        string = _beautify(filtered)
        if filename and log_str:
            with open(filename, "a") as f:
                f.write(f"{log_str}\n")
        return string

    real_serdes_list = [int(i.strip()) for i in serdes.split(",")] if serdes else []
    kw = {"filename": filename, "keep": keep, "serdes": real_serdes_list}
    storage: CmdContext = ctx.obj
    storage.set_loop_coro(log, kw)
    return ""


# --------------------------
# command: strict
# --------------------------
@anlt.command(cls=cb.XenaCommand)
@ac.option("--on/--off", type=ac.BOOL, help=h.HELP_STRICT_ON, default=True)
@ac.pass_context
async def strict(context: ac.Context, on: bool) -> str:
    """
    ANLT strict mode ignores errored frames.
    """
    storage: CmdContext = context.obj

    port_obj = storage.retrieve_port()
    await anlt_utils.anlt_strict(port_obj, on)
    return format_strict(storage, on)

# --------------------------
# command: log-ctrl
# --------------------------
@anlt.command(cls=cb.XenaCommand)
@ac.option("-D/-d", "--debug/--no-debug", type=ac.BOOL, help=h.HELP_LOG_CONTROL_DEBUG_ON, default=True)
@ac.option("-A/-a", "--an-trace/--no-an-trace", type=ac.BOOL, help=h.HELP_LOG_CONTROL_AN_TRACE_ON, default=True)
@ac.option("-L/-l", "--lt-trace/--no-lt-trace", type=ac.BOOL, help=h.HELP_LOG_CONTROL_LT_TRACE_ON, default=True)
@ac.option("-G/-g", "--alg-trace/--no-alg-trace", type=ac.BOOL, help=h.HELP_LOG_CONTROL_ALG_TRACE_ON, default=True)
@ac.option("-P/-p", "--fsm-port/--no-fsm-port", type=ac.BOOL, help=h.HELP_LOG_CONTROL_FSM_PORT_ON, default=False)
@ac.option("-N/-n", "--fsm-an/--no-fsm-an", type=ac.BOOL, help=h.HELP_LOG_CONTROL_FSM_AN_ON, default=True)
@ac.option("-M/-m", "--fsm-an-stimuli/--no-fsm-an-stimuli", type=ac.BOOL, help=h.HELP_LOG_CONTROL_FSM_AN_STIMULI_ON, default=False)
@ac.option("-T/-t", "--fsm-lt/--no-fsm-lt", type=ac.BOOL, help=h.HELP_LOG_CONTROL_FSM_LT_ON, default=True)
@ac.option("-C/-c", "--fsm-lt-coeff/--no-fsm-lt-coeff", type=ac.BOOL, help=h.HELP_LOG_CONTROL_FSM_LT_COEFF_ON, default=False)
@ac.option("-S/-s", "--fsm-lt-stimuli/--no-fsm-lt-stimuli", type=ac.BOOL, help=h.HELP_LOG_CONTROL_FSM_LT_STIMULI_ON, default=False)
@ac.option("-Z/-z", "--fsm-lt-alg0/--no-fsm-lt-alg0", type=ac.BOOL, help=h.HELP_LOG_CONTROL_FSM_LT_ALG0_ON, default=True)
@ac.option("-O/-o", "--fsm-lt-algn1/--no-fsm-lt-algn1", type=ac.BOOL, help=h.HELP_LOG_CONTROL_FSM_LT_ALGN1_ON, default=True)
@ac.pass_context
async def log_ctrl(
    context: ac.Context, 
    debug: bool, 
    an_trace: bool, 
    lt_trace: bool, 
    alg_trace: bool,
    fsm_port: bool,
    fsm_an: bool,
    fsm_an_stimuli: bool,
    fsm_lt: bool,
    fsm_lt_coeff: bool,
    fsm_lt_stimuli: bool,
    fsm_lt_alg0: bool,
    fsm_lt_algn1: bool,
    ) -> str:
    """
    Control what should be logged in ANLT by xenaserver
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    types = []
    if debug:
        types.append(AnLtLogControl.LOG_TYPE_DEBUG)
    if an_trace:
        types.append(AnLtLogControl.LOG_TYPE_AN_TRACE)
    if lt_trace:
        types.append(AnLtLogControl.LOG_TYPE_LT_TRACE)
    if alg_trace:
        types.append(AnLtLogControl.LOG_TYPE_ALG_TRACE)
    if fsm_port:
        types.append(AnLtLogControl.LOG_TYPE_FSM_PORT)
    if fsm_an:
        types.append(AnLtLogControl.LOG_TYPE_FSM_ANEG)
    if fsm_an_stimuli:
        types.append(AnLtLogControl.LOG_TYPE_FSM_ANEG_STIMULI)
    if fsm_lt:
        types.append(AnLtLogControl.LOG_TYPE_FSM_LT)
    if fsm_lt_coeff:
        types.append(AnLtLogControl.LOG_TYPE_FSM_LT_COEFF)
    if fsm_lt_stimuli:
        types.append(AnLtLogControl.LOG_TYPE_FSM_LT_STIMULI)
    if fsm_lt_alg0:
        types.append(AnLtLogControl.LOG_TYPE_FSM_LT_ALG0)
    if fsm_lt_algn1:
        types.append(AnLtLogControl.LOG_TYPE_FSM_LT_ALG1)


    await anlt_utils.anlt_log_control(port_obj, types)
    return format_log_control(
        storage, 
        debug, 
        an_trace, 
        lt_trace, 
        alg_trace,
        fsm_port,
        fsm_an,
        fsm_an_stimuli,
        fsm_lt,
        fsm_lt_coeff,
        fsm_lt_stimuli,
        fsm_lt_alg0,
        fsm_lt_algn1
        )


import asyncssh
import asyncio
import asyncclick
import sys
import os
import typing as t

grandpa = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
if grandpa not in sys.path:
    sys.path.append(grandpa)

from xoa_utils import __version__ as XOA_UTILS_VERSION  # noqa: E402
from xoa_utils.clis import ReadConfig  # noqa: E402
from xoa_utils.hub import Hub  # noqa: E402
from xoa_utils.ssh_server import XenaSSHServer  # noqa: E402
from xoa_utils.cmds import CmdWorker  # noqa: E402


class XenaSSHCLIHandle:
    def __init__(self, config: ReadConfig) -> None:
        self.config = config

    async def handle_client(self, process: asyncssh.SSHServerProcess) -> None:
        out = process.stdout
        setattr(out, "flush", out._chan._flush_send_buf)
        # patch the flush() method since it doesn't exist.
        asyncclick.echo(
            f"Hello {process.get_extra_info('username')}, welcome to Xena OpenAutomation Utilities server ({XOA_UTILS_VERSION})",
            file=t.cast(t.IO, out),
        )
        worker = CmdWorker(process)
        await worker.run(self.config)


async def start_server(config: ReadConfig) -> None:
    Hub.check_hub_process(config)
    print(
        f"(PID: {os.getpid()}) Xena OpenAutomation Utilities server ({XOA_UTILS_VERSION}) running on 0.0.0.0:{config.conn_port}"
    )
    server = await asyncssh.create_server(
        XenaSSHServer,
        "0.0.0.0",
        config.conn_port,
        server_host_keys=[config.conn_host_keys],
        process_factory=XenaSSHCLIHandle(config).handle_client,
    )
    await server.wait_closed()


async def xoa_utility() -> None:
    argv = (sys.argv[1],) if len(sys.argv) >= 2 else tuple()
    config = ReadConfig(*argv)
    try:
        await start_server(config)
    except (OSError, asyncssh.Error) as exc:
        sys.exit(f"Error starting server: <{type(str(exc))}> {exc}")
    except KeyboardInterrupt:
        os.remove(config.hub_pid_path)

def main() -> None:
    asyncio.run(xoa_utility())

if __name__ == "__main__":
    main()



from __future__ import annotations
import asyncclick as ac
from xoa_utils.clicks import click_backend as cb
from xoa_driver.hlfuncs import anlt as anlt_utils
from xoa_driver import enums
from xoa_utils.clis import (
    format_lt_algorithm,
    format_lt_config,
    format_lt_im,
    format_lt_inc_dec,
    format_lt_encoding,
    format_lt_preset,
    format_lt_trained,
    format_txtap_get,
    format_txtap_set,
    format_lt_status,
)
from xoa_utils.clicks.click_commands.group import xoa_util
from xoa_utils.clicks import click_help as h
from xoa_utils.cmds import CmdContext


# --------------------------
# command: lt
# --------------------------
@xoa_util.group(cls=cb.XenaGroup)
def lt():
    """
    Commands for Link Training.
    """


# **************************
# Type: Config
# **************************
# **************************
# sub-command: lt config
# **************************
@lt.command(cls=cb.XenaCommand, name="config")
@ac.option(
    "--mode",
    type=ac.Choice(["interactive", "auto"]),
    help=h.HELP_LT_CONFIG_MODE,
    default="auto",
)
@ac.option("--on/--off", type=ac.BOOL, help=h.HELP_LT_CONFIG_ON, default=True)
@ac.option(
    "--preset0",
    type=ac.Choice(["standard", "existing"]),
    help=h.HELP_LT_CONFIG_PRESET0,
    default="standard",
)
@ac.pass_context
async def lt_config(context: ac.Context, mode: str, on: bool, preset0: str) -> str:
    """
    Configure LT for the working port.
    """
    storage: CmdContext = context.obj
    storage.retrieve_port()
    storage.store_should_do_lt(on)
    storage.store_lt_preset0(preset0)
    storage.store_lt_interactive(True if mode == "interactive" else False)
    return format_lt_config(storage)


# **************************
# sub-command: lt im
# **************************
@lt.command(cls=cb.XenaCommand, name="im")
@ac.argument("serdes", type=ac.INT)
@ac.argument("encoding", type=ac.Choice(["nrz", "pam4", "pam4pre"]))
@ac.pass_context
async def lt_im(context: ac.Context, serdes: int, encoding: str) -> str:
    """
    Set initial modulation for the specified serdes.

        <SERDES>: Specifies the transceiver serdes index.

        <ENCODING>: Specifies the initial modulation. Allowed values: nrz | pam4 | pam4pre
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    storage.validate_current_serdes(serdes)
    storage.store_lt_initial_mod(serdes, encoding)
    status_dic = await anlt_utils.lt_im_status(port_obj)
    return format_lt_im(status_dic, storage, serdes)


# **************************
# sub-command: lt alg
# **************************
@lt.command(cls=cb.XenaCommand, name="alg")
@ac.argument("serdes", type=ac.INT)
@ac.argument("algorithm", type=ac.Choice(["alg0", "algn1"]))
@ac.pass_context
async def lt_algorithm(context: ac.Context, serdes: int, algorithm: str) -> str:
    """
    Set the link training algorithm for the specified serdes.

        <SERDES>: Specifies the transceiver serdes index.

        <ALGORITHM>: Specifies the algorithm. Allowed values: alg0 | algn1
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    storage.validate_current_serdes(serdes)
    storage.store_lt_algorithm(serdes, algorithm)
    status_dic = await anlt_utils.lt_algorithm_status(port_obj)
    return format_lt_algorithm(status_dic, storage, serdes)


# **************************
# Type: Control
# **************************
# **************************
# sub-command: lt inc
# **************************
@lt.command(cls=cb.XenaCommand, name="inc")
@ac.argument("serdes", type=ac.INT)
@ac.argument("emphasis", type=ac.Choice(["pre3", "pre2", "pre", "main", "post"]))
@ac.pass_context
async def lt_inc(context: ac.Context, serdes: int, emphasis: str) -> str:
    """
    Request the remote port's serdes to increase (+) an emphasis by 1.

        <SERDES>: Specifies the transceiver serdes index.

        <EMPHASIS>: The emphasis (coefficient) of the link partner. Allowed values: pre3 | pre2 | pre | main | post
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    storage.validate_current_serdes(serdes)
    resp = await anlt_utils.lt_coeff_inc(
        port_obj, serdes, enums.LinkTrainCoeffs[emphasis.upper()]
    )
    return format_lt_inc_dec(storage, serdes, emphasis, True, resp.name)


# **************************
# sub-command: lt dec
# **************************
@lt.command(cls=cb.XenaCommand, name="dec")
@ac.argument("serdes", type=ac.INT)
@ac.argument("emphasis", type=ac.Choice(["pre3", "pre2", "pre", "main", "post"]))
@ac.pass_context
async def lt_dec(context: ac.Context, serdes: int, emphasis: str) -> str:
    """
    Request the remote port's serdes to decrease (-) an emphasis by 1.

        <SERDES>: The serdes index.

        <EMPHASIS>: The emphasis (coefficient) of the link partner. Allowed values: pre3 | pre2 | pre | main | post
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    storage.validate_current_serdes(serdes)
    resp = await anlt_utils.lt_coeff_dec(
        port_obj, serdes, enums.LinkTrainCoeffs[emphasis.upper()]
    )
    return format_lt_inc_dec(storage, serdes, emphasis, False, resp.name)


# **************************
# sub-command: lt encoding
# **************************
@lt.command(cls=cb.XenaCommand, name="encoding")
@ac.argument("serdes", type=ac.INT)
@ac.argument("encoding", type=ac.Choice(["nrz", "pam4", "pam4pre"]))
@ac.pass_context
async def lt_encoding(context: ac.Context, serdes: int, encoding: str) -> str:
    """
    Request the remote port's serdes to use the specified encoding.

        <SERDES>: The serdes index.

        <ENCODING>: Specifies the encoding. Allowed values: nrz | pam4 | pam4pre
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    storage.validate_current_serdes(serdes)
    e = enums.LinkTrainEncoding[
        {"pam4pre": "PAM4_WITH_PRECODING"}.get(encoding, encoding).upper()
    ]
    resp = await anlt_utils.lt_encoding(port_obj, serdes, e)
    return format_lt_encoding(storage, serdes, encoding, resp.name)


# **************************
# sub-command: lt preset
# **************************
@lt.command(cls=cb.XenaCommand, name="preset")
@ac.argument("serdes", type=ac.INT)
@ac.argument("preset", type=ac.IntRange(1, 5))
@ac.pass_context
async def lt_preset(context: ac.Context, serdes: int, preset: int) -> str:
    """
    Request the remote port's serdes to use the preset.

        <SERDES>: The serdes index.

        <PRESET>: Specifies the preset index. Allowed values: 1 | 2 | 3 | 4 | 5
    """
    preset = max(preset - 1, 0)
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    storage.validate_current_serdes(serdes)
    resp = await anlt_utils.lt_preset(port_obj, serdes, enums.LinkTrainPresets(preset))
    return format_lt_preset(storage, serdes, preset, resp.name)


# **************************
# sub-command: lt trained
# **************************
@lt.command(cls=cb.XenaCommand, name="trained")
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def lt_trained(context: ac.Context, serdes: int) -> str:
    """
    Announce the serdes is trained to the remote port.

        <SERDES>: The serdes index.
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    storage.validate_current_serdes(serdes)
    resp = await anlt_utils.lt_trained(port_obj, serdes)
    return format_lt_trained(storage, serdes, resp.name)


# **************************
# sub-command: lt txtapget
# **************************
@lt.command(cls=cb.XenaCommand, name="txtapget")
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def lt_txtapget(context: ac.Context, serdes: int) -> str:
    """
    Read the tap values of the specified serdes of the working port.

        <SERDES>: The serdes index.
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    storage.validate_current_serdes(serdes)
    dic = await anlt_utils.txtap_get(port_obj, serdes)
    return format_txtap_get(serdes, dic)


# **************************
# sub-command: lt txtapset
# **************************
@lt.command(cls=cb.XenaCommand, name="txtapset")
@ac.argument("serdes", type=ac.INT)
@ac.argument("pre3", type=ac.INT)
@ac.argument("pre2", type=ac.INT)
@ac.argument("pre", type=ac.INT)
@ac.argument("main", type=ac.INT)
@ac.argument("post", type=ac.INT)
@ac.pass_context
async def lt_txtapset(
    context: ac.Context,
    serdes: int,
    pre3: int,
    pre2: int,
    pre: int,
    main: int,
    post: int,
) -> str:
    """
    Write the tap values of the specified serdes of the working port.

        <SERDES>: The serdes index.

        <PRE3>: Specifies c(-3) value of the tap.

        <PRE2>: Specifies c(-2) value of the tap.

        <PRE> : Specifies c(-1) value of the tap.

        <MAIN>: Specifies c(0) value of the tap.

        <POST>: Specifies c(1) value of the tap.
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    storage.validate_current_serdes(serdes)
    await anlt_utils.txtap_set(port_obj, serdes, pre3, pre2, pre, main, post)
    return format_txtap_set(serdes, pre3, pre2, pre, main, post)


# **************************
# sub-command: lt status
# **************************
@lt.command(cls=cb.XenaCommand, name="status")
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def lt_status(context: ac.Context, serdes: int) -> str:
    """
    Show the link training status of the serdes.

        <SERDES>: The serdes index.
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    storage.validate_current_serdes(serdes)
    dic = await anlt_utils.lt_status(port_obj, serdes)
    return format_lt_status(dic)


# ******************************
# sub-command: lt txtap-autotune
# ******************************
@lt.command(cls=cb.XenaCommand, name="txtap-autotune")
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def lt_txtap_autotune(
    context: ac.Context,
    serdes: int,
) -> str:
    """
    Auto tune the tx tap values of the specified serdes of the working port.

        <SERDES>: The serdes index.

    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    storage.validate_current_serdes(serdes)
    await anlt_utils.txtap_autotune(port_obj, serdes)
    dic = await anlt_utils.txtap_get(port_obj, serdes)
    return format_txtap_get(serdes, dic)


from __future__ import annotations
import typing as t
from xoa_utils.cmds import CmdContext
from xoa_utils.clicks.click_commands import an
from xoa_utils.clicks.click_commands import lt
from xoa_utils.clicks.click_commands import management
from xoa_utils.clicks.click_commands import anlt
from xoa_utils.clicks.click_commands import debug
from xoa_utils.clicks.click_commands.group import xoa_util


async def cmd_main(context: CmdContext, cmd_str: str) -> t.Any:
    context.clear_error()
    args = cmd_str.split()
    result = await xoa_util.main(args=args, standalone_mode=False, obj=context)
    return result



from __future__ import annotations
import asyncclick as ac
import typing as t

if t.TYPE_CHECKING:
    from xoa_utils.cmds import CmdContext


class XenaGroup(ac.Group):
    def __init__(self, *args: t.Any, **attrs: dict) -> None:
        attrs.setdefault("context_settings", {})
        attrs["context_settings"].update({"help_option_names": ["-h", "--help"]})
        super().__init__(*args, **attrs)

    def get_help(self, ctx: ac.Context) -> str:
        e = super().get_help(ctx)
        e = e.replace("python -m entry ", "")
        storage: CmdContext = ctx.obj
        storage.set_error(f"{e}\n")
        return e


class XenaCommand(ac.Command):
    def __init__(self, *args: t.Any, **attrs: dict) -> None:
        attrs.setdefault("context_settings", {})
        attrs["context_settings"].update({"help_option_names": ["-h", "--help"]})
        super().__init__(*args, **attrs)

    def get_help(self, ctx: ac.Context) -> str:
        e = super().get_help(ctx)
        e = e.replace("python -m entry ", "")
        storage: CmdContext = ctx.obj
        storage.set_error(f"{e}\n")
        return e



from __future__ import annotations
import asyncclick as ac
from xoa_driver.hlfuncs import anlt as anlt_utils
from xoa_utils.clicks import click_backend as cb
from xoa_utils.clis import format_an_status, format_an_config
from xoa_utils.clicks.click_commands.group import xoa_util
from xoa_utils.clicks import click_help as h
from xoa_utils.cmds import CmdContext


# --------------------------
# command: an
# --------------------------
@xoa_util.group(cls=cb.XenaGroup)
def an():
    """
    Commands for Auto-Negotiation.
    """


# **************************
# Type: Config
# **************************
# **************************
# sub-command: an config
# **************************
@an.command(cls=cb.XenaCommand, name="config")
@ac.option("--on/--off", type=ac.BOOL, help=h.HELP_AN_CONFIG_ON, default=True)
@ac.option(
    "--loopback/--no-loopback",
    type=ac.BOOL,
    help=h.HELP_AN_CONFIG_LOOPBACK,
    default=False,
)
@ac.pass_context
async def an_config(context: ac.Context, on: bool, loopback: bool) -> str:
    """
    Configure AN of the working port.
    """
    storage: CmdContext = context.obj
    storage.retrieve_port()
    storage.store_an_allow_loopback(loopback)
    storage.store_should_do_an(on)
    return format_an_config(storage)


# **************************
# Type: Statistics
# **************************
# **************************
# sub-command: an status
# **************************
@an.command(cls=cb.XenaCommand, name="status")
@ac.pass_context
async def an_status(context: ac.Context) -> str:
    """
    Show AN status of the working port.
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    status_dic = await anlt_utils.autoneg_status(port_obj)
    return format_an_status(status_dic)



from __future__ import annotations
import asyncclick as ac
import typing as t
import json
from xoa_utils.clicks.click_commands.group import xoa_util
from xoa_utils.clicks import click_backend as cb
from xoa_utils.cmds import CmdContext
from xoa_driver.hlfuncs import (
    anlt_ll_debug as debug_utils,
)
import time
import json
import csv
from xoa_utils.clis import (
    format_debug_init
)


@xoa_util.group(cls=cb.XenaGroup)
def debug():
    """
    Commands for debugging.\n
    """


# --------------------------
# command: init
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def init(context: ac.Context, serdes: int) -> str:
    """
    Debug: Initialize debug

        <SERDES>: Serdes index.
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    inf = await debug_utils.init(port_obj, serdes)
    storage.store_anlt_low(serdes, inf)
    inf_dic = {
        "base": inf.base,
        "rx_gtm_base": inf.rx_gtm_base,
        "tx_gtm_base": inf.tx_gtm_base,
        "rx_serdes": inf.rx_serdes,
        "tx_serdes": inf.tx_serdes
    }
    return format_debug_init(inf_dic)


async def _help_get(func: t.Callable, context: ac.Context, serdes: int) -> str:
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    or_inf = inf = storage.retrieve_anlt_low()
    if storage.retrieve_anlt_serdes() == serdes or or_inf is not None:
        inf = or_inf
    else:
        inf = await debug_utils.init(port_obj, serdes)
        storage.store_anlt_low(serdes, inf)
    return str(await func(port_obj, serdes, inf=inf))


async def _help_set(
    func: t.Callable, context: ac.Context, serdes: int, value: int
) -> str:
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    or_inf = inf = storage.retrieve_anlt_low()
    if storage.retrieve_anlt_serdes() == serdes or or_inf is not None:
        inf = or_inf
    else:
        inf = await debug_utils.init(port_obj, serdes)
        storage.store_anlt_low(serdes, inf)
    await func(port_obj, serdes, value=value, inf=inf)
    return ""


# --------------------------
# command: serdes-reset
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def serdes_reset(context: ac.Context, serdes: int) -> str:
    """
    Debug: Reset serdes

        <SERDES>: Serdes index.
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    or_inf = inf = storage.retrieve_anlt_low()
    if storage.retrieve_anlt_serdes() == serdes or or_inf is not None:
        inf = or_inf
    else:
        inf = await debug_utils.init(port_obj, serdes)
        storage.store_anlt_low(serdes, inf)
    await debug_utils.serdes_reset(port_obj, serdes, inf=inf)
    return ""


# --------------------------
# command: mode-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def mode_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Get mode

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.mode_get, context, serdes)


# --------------------------
# command: mode-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.argument("value", type=ac.INT)
@ac.pass_context
async def mode_set(context: ac.Context, serdes: int, value: int) -> str:
    """
    Debug: Set mode

        <SERDES>: Serdes index.

        <VALUE>: Value.
    """

    return await _help_set(debug_utils.mode_set, context, serdes, value)


# --------------------------
# command: an-tx-config-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def an_tx_config_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read AN TX config

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.an_tx_config_get, context, serdes)


# --------------------------
# command: an-tx-config-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.argument("value", type=ac.INT)
@ac.pass_context
async def an_tx_config_set(context: ac.Context, serdes: int, value: int) -> str:
    """
    Debug: Write AN TX config

        <SERDES>: Serdes index.

        <VALUE>: Value.
    """
    return await _help_set(debug_utils.an_tx_config_set, context, serdes, value)


# --------------------------
# command: an-rx-config-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def an_rx_config_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read AN RX config

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.an_rx_config_get, context, serdes)


# --------------------------
# command: an-rx-config-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.argument("value", type=ac.INT)
@ac.pass_context
async def an_rx_config_set(context: ac.Context, serdes: int, value: int) -> str:
    """
    Debug: Write AN RX config

        <SERDES>: Serdes index.

        <VALUE>: Value.
    """
    return await _help_set(debug_utils.an_rx_config_set, context, serdes, value)


# --------------------------
# command: an-rx-dme-mv-range-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def an_rx_dme_mv_range_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read AN MV range

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.an_rx_dme_mv_range_get, context, serdes)

# --------------------------
# command: an-rx-dme-mv-range-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.argument("value", type=ac.INT)
@ac.pass_context
async def an_rx_dme_mv_range_set(context: ac.Context, serdes: int, value: int) -> str:
    """
    Debug: Write AN MV range

        <SERDES>: Serdes index.

        <VALUE>: Value.
    """
    return await _help_set(debug_utils.an_rx_dme_mv_range_set, context, serdes, value)

# --------------------------
# command: an-rx-dme-bit-range-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def an_rx_dme_bit_range_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read AN TX config of the serdes.

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.an_rx_dme_bit_range_get, context, serdes)

# --------------------------
# command: an-rx-dme-bit-range-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.argument("value", type=ac.INT)
@ac.pass_context
async def an_rx_dme_bit_range_set(context: ac.Context, serdes: int, value:int) -> str:
    """
    Debug: Write AN MV range

        <SERDES>: Serdes index.

        <VALUE>: Value.
    """
    return await _help_set(debug_utils.an_rx_dme_bit_range_set, context, serdes, value)

# --------------------------
# command: an-status
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def an_status(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read AN status

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.an_status, context, serdes)


# --------------------------
# command: an-tx-page0-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def an_tx_page0_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read AN TX page0

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.an_tx_page0_get, context, serdes)


# --------------------------
# command: an-tx-page0-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.argument("value", type=ac.INT)
@ac.pass_context
async def an_tx_page0_set(context: ac.Context, serdes: int, value: int) -> str:
    """
    Debug: Write AN PAGE (page0, page1)


        <SERDES>: Serdes index.

        <VALUE>: Value.
    """
    return await _help_set(debug_utils.an_tx_page0_set, context, serdes, value)


# --------------------------
# command: an-tx-page1-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def an_tx_page1_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read AN TX page1

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.an_tx_page1_get, context, serdes)


# --------------------------
# command: an-tx-page1-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.argument("value", type=ac.INT)
@ac.pass_context
async def an_tx_page1_set(context: ac.Context, serdes: int, value: int) -> str:
    """
    Debug: Set page1 of AN page


        <SERDES>: Serdes index.

        <VALUE>: Value.
    """
    return await _help_set(debug_utils.an_tx_page1_set, context, serdes, value)


# --------------------------
# command: an-rx-page0-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def an_rx_page0_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read AN RX page0

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.an_rx_page0_get, context, serdes)


# --------------------------
# command: an-rx-page1-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def an_rx_page1_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read AN RX page1

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.an_rx_page1_get, context, serdes)


# --------------------------
# command: lt-tx-config-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def lt_tx_config_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read LT TX config

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.lt_tx_config_get, context, serdes)


# --------------------------
# command: lt-tx-config-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.argument("value", type=ac.INT)
@ac.pass_context
async def lt_tx_config_set(context: ac.Context, serdes: int, value: int) -> str:
    """
    Debug: Write LT TX config

        <SERDES>: Serdes index.

        <VALUE>: Value.
    """
    return await _help_set(debug_utils.lt_tx_config_set, context, serdes, value)


# --------------------------
# command: lt-rx-config-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def lt_rx_config_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read LT RX config

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.lt_rx_config_get, context, serdes)


# --------------------------
# command: lt-rx-config-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.argument("value", type=ac.INT)
@ac.pass_context
async def lt_rx_config_set(context: ac.Context, serdes: int, value: int) -> str:
    """
    Debug: Write LT RX config

        <SERDES>: Serdes index.

        <VALUE>: Value.
    """
    return await _help_set(debug_utils.lt_rx_config_set, context, serdes, value)


# --------------------------
# command: lt-tx-tf-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def lt_tx_tf_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read LT TX Training Frames

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.lt_tx_tf_get, context, serdes)


# --------------------------
# command: lt-tx-tf-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.argument("value", type=ac.INT)
@ac.pass_context
async def lt_tx_tf_set(context: ac.Context, serdes: int, value: int) -> str:
    """
    Debug: Write LT TX Test Frame

        <SERDES>: Serdes index.

        <VALUE>: Value.
    """
    return await _help_set(debug_utils.lt_tx_tf_set, context, serdes, value)


# --------------------------
# command: lt-rx-tf-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def lt_rx_tf_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read LT RX Test Frame

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.lt_rx_tf_get, context, serdes)


# --------------------------
# command: lt-rx-error-stat0-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def lt_rx_error_stat0_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read LT RX Error Stat0

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.lt_rx_error_stat0_get, context, serdes)


# --------------------------
# command: lt-rx-error-stat1-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def lt_rx_error_stat1_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read LT RX Error Stat1

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.lt_rx_error_stat1_get, context, serdes)


# --------------------------
# command: lt-status
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def lt_status(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read LT status

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.lt_status, context, serdes)


# --------------------------
# command: lt-prbs-ber
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def lt_prbs_ber(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read LT BRPS-13 BER

        <SERDES>: Serdes index.
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    or_inf = inf = storage.retrieve_anlt_low()
    if storage.retrieve_anlt_serdes() == serdes or or_inf is not None:
        inf = or_inf
    else:
        inf = await debug_utils.init(port_obj, serdes)
        storage.store_anlt_low(serdes, inf)
    dic = await debug_utils.lt_prbs(port_obj, serdes, inf=inf)
    return json.dumps(dic, indent=2)


# --------------------------
# command: xla-config-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def xla_config_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read Xena Logic Analyzer config

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.xla_config_get, context, serdes)


# --------------------------
# command: xla-config-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.argument("value", type=ac.INT)
@ac.pass_context
async def xla_config_set(context: ac.Context, serdes: int, value: int) -> str:
    """
    Debug: Write Xena Logic Analyzer config

        <SERDES>: Serdes index.

        <VALUE>: Value.
    """
    await _help_set(debug_utils.xla_config_set, context, serdes, value)
    return ""


# --------------------------
# command: xla-trig-mask-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def xla_trig_mask_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read Xena Logic Analyzer trigger mask

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.xla_trig_mask_get, context, serdes)


# --------------------------
# command: xla-trig-mask-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.argument("value", type=ac.INT)
@ac.pass_context
async def xla_trig_mask_set(context: ac.Context, serdes: int, value: int) -> str:
    """
    Debug: Write Xena Logic Analyzer trigger mask

        <SERDES>: Serdes index.

        <VALUE>: Value.
    """
    await _help_set(debug_utils.xla_trig_mask_set, context, serdes, value)
    return ""


# --------------------------
# command: xla-status-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def xla_status_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read Xena Logic Analyzer status

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.xla_status_get, context, serdes)


# --------------------------
# command: xla-rd-addr-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def xla_rd_addr_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read Xena Logic Analyzer RD Address

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.xla_rd_addr_get, context, serdes)


# --------------------------
# command: xla-rd-addr-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.argument("value", type=ac.INT)
@ac.pass_context
async def xla_rd_addr_set(context: ac.Context, serdes: int, value: int) -> str:
    """
    Debug: Write Xena Logic Analyzer RD Address

        <SERDES>: Serdes index.

        <VALUE>: Value.
    """
    await _help_set(debug_utils.xla_rd_addr_set, context, serdes, value)
    return ""


# --------------------------
# command: xla-rd-page-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def xla_rd_page_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read Xena Logic Analyzer RD Page

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.xla_rd_page_get, context, serdes)


# --------------------------
# command: xla-rd-page-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.argument("value", type=ac.INT)
@ac.pass_context
async def xla_rd_page_set(context: ac.Context, serdes: int, value: int) -> str:
    """
    Debug: Write Xena Logic Analyzer RD Page

        <SERDES>: Serdes index.

        <VALUE>: Value.
    """
    await _help_set(debug_utils.xla_rd_page_set, context, serdes, value)
    return ""


# --------------------------
# command: xla-rd-data-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.pass_context
async def xla_rd_data_get(context: ac.Context, serdes: int) -> str:
    """
    Debug: Read Xena Logic Analyzer RD Data

        <SERDES>: Serdes index.
    """
    return await _help_get(debug_utils.xla_rd_data_get, context, serdes)


# --------------------------
# command: xla-dump
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.option("-f", "--filename", type=ac.STRING, default="")
@ac.pass_context
async def xla_dump(context: ac.Context, serdes: int, filename: str) -> str:
    """
    Debug: Show the Xena Logic Analyzer dump

        <SERDES>: Serdes index.
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    or_inf = inf = storage.retrieve_anlt_low()
    if storage.retrieve_anlt_serdes() == serdes or or_inf is not None:
        inf = or_inf
    else:
        inf = await debug_utils.init(port_obj, serdes)
        storage.store_anlt_low(serdes, inf)
    ret_dict = await debug_utils.xla_dump(port_obj, serdes, inf=inf)
    fieldnames = []
    for key in ret_dict.keys():
        fieldnames.append(key)
    if filename:
        with open(filename, "w+") as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames = fieldnames)
            writer.writeheader()
            writer.writerow(ret_dict)
        return "Result stored in csv file"
    else:
        return json.dumps(ret_dict, indent=2)


# --------------------------
# command: xla-trig-n-dump
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("serdes", type=ac.INT)
@ac.option("-m", "--mask", type=ac.STRING, default="0x00000FF0")
@ac.option("-o", "--window-offset", type=ac.STRING, default="0x0080")
@ac.option("-s", "--trigger-select", type=ac.STRING, default="0x0001")
@ac.option("-f", "--filename", type=ac.STRING, default="xla_dump.csv")
@ac.pass_context
async def xla_trig_n_dump(context: ac.Context, serdes: int, mask: str, window_offset: str, trigger_select: str, filename: str) -> str:
    """
    Debug: Trigger and dump the Xena Logic Analyzer data

        <SERDES>: Serdes index.
    """
    if not mask.startswith("0x"):
        return "Missing 0x for --mask option"
    if not window_offset.startswith("0x"):
        return "Missing 0x for --window-offset option"
    if not trigger_select.startswith("0x"):
        return "Missing 0x for --trigger-select option"
    
    # Set the XLA trigger mask and verify the result
    try:
        mask_value = int(mask, 16)
        await _help_set(debug_utils.xla_trig_mask_set, context, serdes, mask_value)
        resp = await _help_get(debug_utils.xla_trig_mask_get, context, serdes)
        if int(resp) != mask_value:
            return f"Setting mask failed. (expecting {mask} but get {resp:#010x})"
    except ValueError:
        return "Invalid hex string for --mask option"
    
    # Config the XLA trigger and verify the result
    try:
        window_offset_value = int(window_offset, 16) << 16
        trigger_select_value = int(trigger_select, 16)
        config_value = window_offset_value + trigger_select_value
        await _help_set(debug_utils.xla_config_set, context, serdes, config_value)
        resp = await _help_get(debug_utils.xla_config_get, context, serdes)
        if int(resp) != window_offset_value:
            return f"Configuring trigger failed. (expecting {window_offset_value:#010x} but get {resp:#010x})"
    except ValueError:
        return "Invalid hex string for --window-offset and/or --trigger-select options"
    
    # Check status and dump data
    xla_status = 0
    for _ in range(10):
        xla_status = await _help_get(debug_utils.xla_status_get, context, serdes)
        if int(xla_status) == 1:
            break
        else:
            time.sleep(1)
    if int(xla_status) == 0:
        return "XLA status time out"
    
    # Dump data
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    or_inf = inf = storage.retrieve_anlt_low()
    if storage.retrieve_anlt_serdes() == serdes or or_inf is not None:
        inf = or_inf
    else:
        inf = await debug_utils.init(port_obj, serdes)
        storage.store_anlt_low(serdes, inf)
    ret_dict = await debug_utils.xla_dump(port_obj, serdes, inf=inf)
    fieldnames = []
    for key in ret_dict.keys():
        fieldnames.append(key)
    if filename:
        with open(filename, "w+") as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames = fieldnames)
            writer.writeheader()
            writer.writerow(ret_dict)
        return "Result stored in csv file"
    else:
        return json.dumps(ret_dict, indent=2)
    

# --------------------------
# command: px-get
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("page", type=ac.INT)
@ac.argument("reg", type=ac.STRING)
@ac.pass_context
async def px_get(context: ac.Context, page: int, reg: str) -> str:
    """
    Debug: Get register value.

        <PAGE_ADDRESS>: Page address in decimal.

        <REG_ADDRESS>: Register address in HEX string.
    """
    if not reg.startswith("0x"):
        return "Missing 0x for <REG_ADDRESS>"
    
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()

    resp = await debug_utils.px_get(port_obj, page_address=page, register_address=int(reg,16))
    if resp[0] == False:
        return f"\033[91mDEAD Error\033[0m: PAGE_ADDRESS: {page}, REG_ADDRESS: {reg}, value: {resp[1]}"
    else:
        return f"PAGE_ADDRESS: {page}, REG_ADDRESS: {reg}, value: {resp[1]}"

# --------------------------
# command: px-set
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.argument("page", type=ac.INT)
@ac.argument("reg", type=ac.STRING)
@ac.argument("value", type=ac.STRING)
@ac.pass_context
async def px_set(context: ac.Context, page: int, reg: str, value: str) -> str:
    """
    Debug: Set register value

        <PAGE_ADDRESS>: Page address in decimal.

        <REG_ADDRESS>: Register address in HEX string.

        <VALUE>: Value to write in HEX string
    """
    
    if not reg.startswith("0x"):
        return "Missing 0x for <REG_ADDRESS>"
    if not value.startswith("0x"):
        return "Missing 0x for <VALUE>"
    
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()

    await debug_utils.px_set(port_obj, page_address=page, register_address=int(reg,16), value=int(value,16))
    resp = await debug_utils.px_get(port_obj, page_address=page, register_address=int(reg,16))
    if resp[0] == False:
        return f"\033[91mFailed\033[0m: PAGE_ADDRESS: {page}, REG_ADDRESS: {reg}, value: {resp[1]}"
    else:
        return f"\033[92mSuccess\033[0m: PAGE_ADDRESS: {page}, REG_ADDRESS: {reg}, value: {resp[1]}"
    

# --------------------------
# command: xla-dump-ctrl
# --------------------------
@debug.command(cls=cb.XenaCommand)
@ac.option("--on/--off", type=ac.BOOL, default=True)
@ac.pass_context
async def xla_dump_ctrl(context: ac.Context, on: bool) -> str:
    """
    Debug: Control XLA auto dump
    """
    storage: CmdContext = context.obj
    port_obj = storage.retrieve_port()
    await debug_utils.xla_dump_ctrl(port_obj, on)
    return f"XLA auto dump {'on' if on else 'off'}"


from __future__ import annotations

from xoa_driver.ports import GenericAnyPort
import typing as t


class ConfigError(Exception):
    def __init__(self) -> None:
        self.name = ""
        self.msg = ""

    def __repr__(self) -> str:
        return self.msg

    def __str__(self) -> str:
        return self.msg


class NotConnectedError(ConfigError):
    def __init__(self) -> None:
        self.name = "Not Connected Error"
        self.msg = "No tester is connected!"


class NoSuchModuleError(ConfigError):
    def __init__(self, module_id: int) -> None:
        self.name = "No Such Module Error"
        self.msg = f"No such module {module_id}!"


class NoSuchIDError(ConfigError):
    def __init__(self, general_id: str) -> None:
        self.name = "No Such ID Error"
        self.msg = f"No such ID {general_id}!"


class NotInStoreError(ConfigError):
    def __init__(self, general_id: str) -> None:
        self.name = "Not in store Error"
        self.msg = f"Not in store {general_id}!"


class NotCorrectSerdesError(ConfigError):
    def __init__(self, port_id: str, serdes: int) -> None:
        self.name = "Not Correct Serdes Error"
        self.msg = f"No serdes {serdes} in port {port_id}!"


class NoSuchPortError(ConfigError):
    def __init__(self, port_id: int) -> None:
        self.name = "No Such Port Error"
        self.msg = f"No such port {port_id}!"


class NoWorkingPort(ConfigError):
    def __init__(self) -> None:
        self.name = "No Working Port Error"
        self.msg = (
            f"There is no working port, please use 'port' command to specify one!"
        )


class NotSupportPcsPmaError(ConfigError):
    def __init__(self, port: GenericAnyPort) -> None:
        module_id, port_id = port.kind.module_id, port.kind.port_id
        self.name = "Not Support Pcs Pma Error"
        self.msg = f"This port {module_id}/{port_id} does not support PCS/PMA!"


class NotSupportAutoNegError(ConfigError):
    def __init__(self, port: GenericAnyPort) -> None:
        module_id, port_id = port.kind.module_id, port.kind.port_id
        self.name = "Not Support Autoneg Error"
        self.msg = f"This port {module_id}/{port_id} does not support auto-negotiation!"


class NotSupportLinkTrainError(ConfigError):
    def __init__(self, port: GenericAnyPort) -> None:
        module_id, port_id = port.kind.module_id, port.kind.port_id
        self.name = "Not Support Link Train Error"
        self.msg = f"This port {module_id}/{port_id} does not support link training!"


class NotRightLaneLengthError(ConfigError):
    def __init__(self, serdes: list[int]) -> None:
        self.name = "Not Right Serdes Length Error"
        self.msg = f"Serdes {serdes} should be length of 4!"


class NotRightLaneValueError(ConfigError):
    def __init__(self, serdes: list[int]) -> None:
        self.name = "Not Right Serdes Value Error"
        self.msg = f"Serdes {serdes} should be a list of 4 integers ranges from 0 to 255!"


# class NotInChoicesError(ConfigError):
#     def __init__(self, string: str, choices: list[str]) -> None:
#         self.name = "Not In Choices Error"
#         self.msg = f"{string} is not in the choices {choices}!"


class CannotConvertError(ConfigError):
    def __init__(self, e: Exception) -> None:
        self.name = "Cannot Convert Error"
        self.msg = f"{e}!"


class NotSameLengthError(ConfigError):
    def __init__(self, a: t.Iterable, b: t.Iterable) -> None:
        self.name = "Zip Error"
        self.msg = f"{a} {b} are not of the same length!"



from __future__ import annotations
import multiprocessing
from multiprocessing import managers
from signal import SIGTERM
import os
import sys
from typing import Optional
import psutil
from xoa_utils.clis import ReadConfig

global_list = []


class HubManager(managers.SyncManager):
    pass


HubManager.register("get_list", lambda: global_list, managers.ListProxy)


class Hub:
    def __init__(self, server: str, run_port: int) -> None:
        self.manager = HubManager(address=(server, run_port), authkey=b"")

    def serve(self) -> None:
        server = self.manager.get_server()
        server.serve_forever()

    @staticmethod
    def serve_hub(server: str, run_port: int, pid_path: str) -> None:
        pid = os.getpid()
        print(f"(PID: {pid}) Starting a new hub.")
        with open(pid_path, "w") as f:
            f.write(f"{os.getpid()}")
        hub = Hub(server, run_port)
        hub.serve()

    @staticmethod
    def kill_hub(pid_path: str, pid: int) -> None:
        if pid == 0:
            return
        try:
            print(f"(PID: {pid}) Killing existing hub.")
            os.kill(int(pid), SIGTERM)
            with open(pid_path, "w"):
                pass
        except Exception as e:
            print(e)

    @staticmethod
    def start_hub(config: Optional[ReadConfig] = None):
        if config is None:
            config = ReadConfig()
        if len(sys.argv) >= 2 and sys.argv[1] == "kill":
            Hub.kill_hub(config.hub_pid_path, config.hub_pid)
        else:
            Hub.kill_hub(config.hub_pid_path, config.hub_pid)
            Hub.serve_hub(
                config.hub_host,
                config.hub_port,
                config.hub_pid_path,
            )

    @staticmethod
    def start_hub_in_process(config: ReadConfig) -> None:
        args = (config,)
        process = multiprocessing.Process(name="Hub", target=Hub.start_hub, args=args)
        process.start()

    @staticmethod
    def check_hub_process(config: ReadConfig) -> None:
        if not config.hub_enabled:
            Hub.kill_hub(config.hub_pid_path, config.hub_pid)
            return
        pid = config.hub_pid
        if not pid:
            Hub.start_hub_in_process(config)
        else:
            try:
                psutil.Process(pid)
            except psutil.NoSuchProcess:
                Hub.start_hub_in_process(config)


if __name__ == "__main__":
    Hub.start_hub()



import asyncssh
import sys


class XenaSSHServer(asyncssh.SSHServer):
    def connection_made(self, conn: "asyncssh.SSHServerConnection"):
        remote_host = conn.get_extra_info("peername")[0]
        print(f"{remote_host}: SSH connection received from {remote_host}.")

    def connection_lost(self, exc: Exception) -> None:
        if exc and str(exc) != "Connection lost":
            print("SSH connection error: " + str(exc), file=sys.stderr)
        else:
            print("SSH connection closed.")

    def begin_auth(self, username: str) -> bool:
        # If the user's password is the empty string, no auth is required
        return False

    def password_auth_supported(self) -> bool:
        return True

    def validate_password(self, username: str, password: str) -> bool:
        return True



HELP_CONNECT_PORT_LIST = "Specifies the ports on the specified device host, \
default to null. Specify a port using the format slot/port, no spaces between. \
e.g. --port_list 0/0,0/1,0/2,0/3. If used,the context will switch to the first \
port in the list after the connection is established.\n"

HELP_CONNECT_RESET = "Removes all port configurations of the ports in --ports after \
reservation, default to --no-reset."

HELP_CONNECT_FORCE = "Breaks port locks established by another user, aka. \
force reservation, default to --force."


HELP_CONNECT_PWD = "The login password of the tester, default to xena. \
e.g. --password xena"


HELP_CONNECT_TCP_PORT = "The TCP port number on the chassis for the client to \
establish a session, default to 22606."

HELP_CONNECT = """
    Connect to a tester for the current session.

        DEVICE TEXT: Specifies the chassis address for connection. You can specify the IP addresses in IPv4 format, or a host name. e.g. 10.10.10.10, demo.xenanetworks.com\n
        USERNAME TEXT: Specifies the name of the user, e.g. xoa or automation

    """

HELP_PORT_RESET = "Removes all port configurations in --port_list after \
reservation, default to --reset"

HELP_PORT_FORCE = "Breaks port locks established by another user, aka. force \
reservation, default to --force"

HELP_EXIT_RELEASE = "Determines whether the ports should be released before exiting, default to --release"

HELP_PORTS_ALL = "Show all ports of the tester, default to --no-all"

HELP_RECOVERY_ON = "Should xenaserver automatically do link recovery when detecting down signal, default to --on."

HELP_AN_CONFIG_ON = "Enable or disable auto-negotiation on the working port, \
default to --on."

HELP_AN_CONFIG_LOOPBACK = "Should loopback be allowed in auto-negotiation, \
default to --no-loopback."

HELP_LT_CONFIG_MODE = "The mode for link training on the working port, default to --mode auto."

HELP_LT_CONFIG_ON = "Enable or disable link training on the working port, default to --on."

HELP_LT_CONFIG_PRESET0 = "Should the Out-of-Sync use standard values (--preset0 standard) or existing tap values (--preset0 existing), default to --preset0 standard."

HELP_ANLT_LOG_FILENAME = "Filename of the log."

HELP_ANLT_LOG_KEEP = "Specifies what types of log entries to keep, default to keep all.\
Allow values: all | an | lt \
all: keep both AN and LT\
an:  keep AN only\
lt:  keep lt only"

HELP_ANLT_LOG_SERDES = "Specifies which serdes of LT logs to keep. If you don't know how many\
serdes the port has, use 'anlt log', default to all serdes."

HELP_STRICT_ON = "Should enable ANLT strict mode, default to --on."

HELP_LOG_CONTROL_DEBUG_ON = "Debug log out, default to --debug, -D"

HELP_LOG_CONTROL_AN_TRACE_ON = "Auto-negotiation trace output, default to --an-trace, -A"

HELP_LOG_CONTROL_LT_TRACE_ON = "Link training algorithm trace, default to --lt-trace, -L"

HELP_LOG_CONTROL_ALG_TRACE_ON = "Link training algorithm trace output, default to --alg-trace, -G"

HELP_LOG_CONTROL_FSM_PORT_ON = "Port state machine transitions output, default to --no-fsm-port, -p"

HELP_LOG_CONTROL_FSM_AN_ON = "Auto-negotiation state machine transitions, default to --fsm-an, -N"

HELP_LOG_CONTROL_FSM_AN_STIMULI_ON = "Auto-negotiation stimuli state machine transitions, default to --no-fsm-an-stimuli, -m"

HELP_LOG_CONTROL_FSM_LT_ON = "Link training state machine transitions, default to --fsm-lt, -T"

HELP_LOG_CONTROL_FSM_LT_COEFF_ON = "Link training coefficient state machine transitions, default to --no-fsm-lt-coeff, -c"

HELP_LOG_CONTROL_FSM_LT_STIMULI_ON = "Link training stimuli state machine transitions, default to --no-fsm-lt-stimuli, -s"

HELP_LOG_CONTROL_FSM_LT_ALG0_ON = "Link training algorithm 0 state machine transitions, default to --fsm-lt-alg0, -Z"

HELP_LOG_CONTROL_FSM_LT_ALGN1_ON = "Link training algorithm -1 state machine transitions, default to --fsm-lt-algn1, -O"


from __future__ import annotations
from asyncclick.shell_completion import ShellComplete, CompletionItem
from asyncclick import BaseCommand
import typing as t
from xoa_utils.clicks import xoa_util
from xoa_utils.clicks import cmd_main
from xoa_utils.hub import HubManager
from xoa_utils.clis import ReadConfig, run_coroutine_as_sync, format_error
import asyncssh as ah
import asyncclick as ac
import os
import asyncio
from xoa_driver import exceptions as driver_ex
from xoa_utils.cmds.cmd_context import CmdContext

if t.TYPE_CHECKING:
    from asyncssh.process import SSHServerProcess
    from asyncssh.editor import SSHLineEditorChannel


class AutoCompleter(ShellComplete):
    """
    Extends ShellComplete. https://click.palletsprojects.com/en/8.1.x/shell-completion/
    """

    def __init__(
        self,
        cli: BaseCommand,
        ctx_args: dict[str, t.Any],
        prog_name: str,
        complete_var: str,
        args_raw: list[str],
    ) -> None:
        super().__init__(cli, ctx_args, prog_name, complete_var)
        self.args_raw = args_raw

    def get_completion_args(self) -> tuple[list[str], str]:
        """Use the self.args to return a tuple of ``args, incomplete``."""
        args = self.args_raw[:-1] if self.args_raw else []
        incomplete = str(self.args_raw[-1]) if self.args_raw else ""
        return args, incomplete

    def format_completion(self, item: CompletionItem) -> str:
        """Format a completion item into the form recognized by the
        shell script.

        :param item: Completion item to format.
        """
        return f"{item.value}"

    async def complete(self) -> str:
        """Produce the completion data to send back to the shell.

        By default this calls :meth:`get_completion_args`, gets the
        completions, then calls :meth:`format_completion` for each
        completion.
        """
        args, incomplete = self.get_completion_args()
        completions = await self.get_completions(args, incomplete)
        out = [self.format_completion(item) for item in completions]
        return "\t".join(out)


async def shell_complete(cli: BaseCommand, args_raw: list[str]) -> str:
    """Perform shell completion for the given CLI program.
    Mimic of :func: asyncclick.shell_completion.shell_complete.

    :param cli: Command being called.

    :return: String after completion
    """
    prog_name = "xoa_utils"
    ctx_args = {}
    completer = AutoCompleter(cli, ctx_args, prog_name, "", args_raw)
    completed = await completer.complete()
    return completed


class CmdWorker:
    def __init__(
        self,
        process: "SSHServerProcess",
        base_prompt: str = "xoa-utils",
    ) -> None:
        self.process: "SSHServerProcess" = process
        self.base_prompt: str = base_prompt
        self.channel: "SSHLineEditorChannel" = self.process.channel  # type: ignore
        self.hub_manager: t.Optional["HubManager"] = None
        self.hub_enable: bool = False
        self.hub_msg_list: list = []
        self.context = CmdContext()
        self.show_prompt = True
        self.register_keys()

    def autocomplete(self, line: str, pos: int) -> tuple[str, int]:
        if not line:
            return line, pos
        args_raw = line.split()
        coro = shell_complete(xoa_util, args_raw)
        completed = str(run_coroutine_as_sync(coro))
        if not completed:
            pass
        elif "\t" in completed:
            self.write(f"{line}\n{completed}\n\n{self.make_prompt()}")
        elif args_raw and completed.startswith(args_raw[-1]):
            new_l = args_raw[:-1] + [completed]
            line = " ".join(new_l)
        else:
            line = completed
        pos = len(line)
        return line, pos

    def register_keys(self) -> None:
        TAB = "\t"
        CTRL_Z = "\x1a"
        self.channel.register_key(TAB, self.autocomplete)
        self.channel.register_key(CTRL_Z, self.stop_coro)

    def stop_coro(self, line: str, pos: int) -> tuple[str, int]:
        self.context.clear_loop_coro()
        self.show_prompt = True
        return line, pos

    def finish(self) -> None:
        self.process.exit(0)

    def write(self, msg: str) -> None:
        self.process.stdout.write(msg)

    def put_hub_record(self, request: str, response: str, success: bool) -> None:
        if self.hub_enable and self.hub_manager:
            self.hub_msg_list.append((os.getpid(), request, response, success))  # type: ignore

    def make_prompt(self, end_prompt: str = ">") -> str:
        return self.context.prompt(self.base_prompt, end_prompt)

    def show_prompts(self, end_prompt: str) -> None:
        if self.show_prompt:
            self.write(f"\n{self.make_prompt(end_prompt)}")
        else:
            self.show_prompt = True

    async def run_interactive(self) -> None:
        self.show_prompts(">")
        request = (await self.process.stdin.readline()).strip()
        response = None
        success = False
        try:
            response = await self.dispatch(request)
            success = True
            if isinstance(response, int):
                response = self.context.get_error()
                success = False
        except ac.UsageError as error:
            response = format_error(error)
            success = False
        except driver_ex.BadStatus as ee:
            response = f"{type(ee).__name__}: Driver BadStatus error.\n"
            success = False
        except Exception as e:
            response = f"{type(e).__name__}: {e}\n"
            success = False
        if response is not None:
            self.write(f"{response}")
            self.put_hub_record(request, response, success)

    async def run_coroutine(self) -> None:
        async_func, kw = self.context.get_loop_coro()
        if async_func is not None:
            try:
                self.show_prompts("!")
                result = await async_func(self.context, **kw)
                if result:
                    self.write(f"{result}\n")
                else:
                    self.show_prompt = False
                await asyncio.sleep(self.context.get_coro_interval())
            except Exception as e:
                self.write(f"{type(e).__name__}: {e}\n")
                self.context.clear_loop_coro()

    async def run(self, config: ReadConfig) -> None:
        self.connect_hub(config)
        while not self.process.stdin.at_eof():
            try:
                if self.context.has_loop_coro():
                    await self.run_coroutine()
                else:
                    await self.run_interactive()
            except ah.TerminalSizeChanged:
                self.show_prompt = False
            except ah.BreakReceived:
                self.finish()

    def connect_hub(self, config: ReadConfig) -> None:
        self.hub_enable = config.hub_enabled
        if config.hub_enabled:
            self.hub_manager = HubManager(
                address=(config.hub_host, config.hub_port), authkey=b""
            )
            self.hub_manager.connect()
            self.hub_msg_list = self.hub_manager.get_list()  # type: ignore

    async def dispatch(self, msg: str) -> str:
        if not msg:
            return ""
        if msg.lower() == "exit":
            await cmd_main(self.context, msg)
            self.finish()
            return ""
        return await cmd_main(self.context, msg)



from __future__ import annotations
import typing as t
from xoa_driver.testers import L23Tester
from xoa_driver.modules import GenericL23Module
from xoa_driver.ports import GenericL23Port
from xoa_utils.exceptions import (
    NotInStoreError,
    NotConnectedError,
    NoSuchIDError,
    NoWorkingPort,
    NotCorrectSerdesError,
)
from xoa_driver.hlfuncs import mgmt as mgmt_utils
from xoa_driver.hlfuncs import anlt_ll_debug as debug_utils
from xoa_driver.enums import NRZPreset, LinkTrainEncoding, LinkTrainAlgorithm
from functools import partialmethod


class ErrorString:
    def __init__(self) -> None:
        self.err_str: str = ""

    def set_val(self, value: str) -> None:
        self.err_str = value

    clear = partialmethod(set_val, "")


class TesterState:
    def __init__(self) -> None:
        self.serial: str = ""
        self.con_info: str = ""
        self.username: str = ""
        self.obj: t.Optional[L23Tester] = None


class ModuleState:
    def __init__(self) -> None:
        self.modules: dict[str, GenericL23Module] = {}
        self.module_str: str = ""


class PortState:
    def __init__(self) -> None:
        self.ports: dict[str, GenericL23Port] = {}
        self.port_str: str = ""
        self.port_serdes_num: dict[str, int] = {}


class ANState:
    def __init__(self) -> None:
        self.allow_loopback: bool = False
        self.do: bool = False


class LTState:
    def __init__(self) -> None:
        self.do: bool = False
        self.preset0: NRZPreset = NRZPreset.NRZ_NO_PRESET
        self.interactive: bool = False
        self.initial_mod: dict[str, LinkTrainEncoding] = {}
        self.algorithm: dict[str, LinkTrainAlgorithm] = {}


class LoopFuncState:
    def __init__(self) -> None:
        self.func: t.Optional[t.Callable] = None
        self.interval: int = 1
        self.kw: dict = {}


class AnltLowState:
    def __init__(self) -> None:
        self.serdes: int = -1
        self.low: t.Optional[debug_utils.AnLtLowLevelInfo] = None


class CmdContext:
    def __init__(self) -> None:
        self.clear_all()

    def clear_all(self) -> None:
        self._tr_state: TesterState = TesterState()
        self._md_state: ModuleState = ModuleState()
        self._pt_state: PortState = PortState()
        self._error: ErrorString = ErrorString()
        self._an_state: ANState = ANState()
        self._lt_state: LTState = LTState()
        self._fn_state: LoopFuncState = LoopFuncState()
        self._anlt_low_state: AnltLowState = AnltLowState()

    def get_coro_interval(self) -> int:
        return self._fn_state.interval

    def has_loop_coro(self) -> bool:
        return self._fn_state.func is not None and callable(self._fn_state.func)

    def get_loop_coro(self) -> tuple[t.Optional[t.Callable], dict]:
        return self._fn_state.func, self._fn_state.kw

    def set_loop_coro(self, coro: t.Optional[t.Callable], dic: dict) -> None:
        self._fn_state.func = coro
        self._fn_state.kw = dic

    clear_loop_coro = partialmethod(set_loop_coro, None, {})

    def prompt(self, base_prompt: str = "", end_prompt: str = ">") -> str:
        s = self.retrieve_tester_serial()
        serial = f"[{s}]" if s else ""
        p = self.retrieve_port_str()
        port_str = f"[{p}]" if p else ""
        return f"{base_prompt}{serial}{port_str} {end_prompt} "

    def store_anlt_low(self, serdes: int, low: debug_utils.AnLtLowLevelInfo) -> None:
        self._anlt_low_state.low = low
        self._anlt_low_state.serdes = serdes

    def store_lt_initial_mod(self, serdes: int, encoding: str) -> None:
        e = LinkTrainEncoding[
            {"pam4pre": "PAM4_WITH_PRECODING"}.get(encoding, encoding).upper()
        ]
        self._lt_state.initial_mod[str(serdes)] = e

    def store_lt_algorithm(self, serdes: int, algorithm: str) -> None:
        e = LinkTrainAlgorithm[algorithm.upper()]
        self._lt_state.algorithm[str(serdes)] = e

    def store_an_allow_loopback(self, do: bool) -> None:
        self._an_state.allow_loopback = do

    def store_should_do_an(self, do: bool) -> None:
        self._an_state.do = do

    def store_should_do_lt(self, do: bool) -> None:
        self._lt_state.do = do

    def store_lt_interactive(self, do: bool) -> None:
        self._lt_state.interactive = do

    def store_lt_preset0(self, preset0: str) -> None:
        if preset0 == "standard":
            self._lt_state.preset0 = NRZPreset.NRZ_NO_PRESET
        else:
            self._lt_state.preset0 = NRZPreset.NRZ_WITH_PRESET

    def store_current_module_str(self, current_module_str: str) -> None:
        if current_module_str not in self._md_state.modules:
            raise NotInStoreError(current_module_str)
        self._md_state.module_str = current_module_str

    def store_current_port_str(self, current_port_str: str) -> None:
        if current_port_str not in self._pt_state.ports:
            raise NotInStoreError(current_port_str)
        self._pt_state.port_str = current_port_str
        self._anlt_low_state = AnltLowState()

    def store_current_tester(
        self, username: str, con_info: str, tester: L23Tester
    ) -> None:
        self.clear_all()
        self._tr_state.con_info = con_info
        self._tr_state.serial = str(tester.info.serial_number)
        self._tr_state.username = username
        self._tr_state.obj = tester

    def get_all_ports(self) -> dict[str, GenericL23Port]:
        return self.obtain_physical_ports("*", False)

    def retrieve_anlt_low(self) -> t.Optional[debug_utils.AnLtLowLevelInfo]:
        return self._anlt_low_state.low

    def retrieve_anlt_serdes(self) -> int:
        return self._anlt_low_state.serdes

    def retrieve_an_enable(self) -> bool:
        return self._an_state.do

    def retrieve_an_loopback(self) -> bool:
        return self._an_state.allow_loopback

    def retrieve_lt_initial_mod(self) -> dict[str, LinkTrainEncoding]:
        return self._lt_state.initial_mod

    def retrieve_lt_initial_mod_serdes(self, serdes: int) -> LinkTrainEncoding:
        if str(serdes) not in self._lt_state.initial_mod:
            raise NotInStoreError(str(serdes))
        return self._lt_state.initial_mod[str(serdes)]

    def retrieve_lt_algorithm(self) -> dict[str, LinkTrainAlgorithm]:
        return self._lt_state.algorithm

    def retrieve_lt_algorithm_serdes(self, serdes: int) -> LinkTrainAlgorithm:
        if str(serdes) not in self._lt_state.algorithm:
            raise NotInStoreError(str(serdes))
        return self._lt_state.algorithm[str(serdes)]

    def retrieve_lt_interactive(self) -> bool:
        return self._lt_state.interactive

    def retrieve_lt_enable(self) -> bool:
        return self._lt_state.do

    def retrieve_lt_preset0(self) -> NRZPreset:
        return self._lt_state.preset0

    def retrieve_ports(self) -> dict[str, GenericL23Port]:
        return self._pt_state.ports

    def retrieve_modules(self) -> dict[str, GenericL23Module]:
        return self._md_state.modules

    def retrieve_tester_username(self) -> str:
        return self._tr_state.username

    def retrieve_tester_con_info(self) -> str:
        return self._tr_state.con_info

    def retrieve_tester_serial(self) -> str:
        return self._tr_state.serial

    def retrieve_tester(self) -> t.Optional[L23Tester]:
        return self._tr_state.obj

    def retrieve_port_str(self) -> str:
        return self._pt_state.port_str

    def store_module(self, exact_module_id: str, module_obj: GenericL23Module) -> None:
        self._md_state.modules[exact_module_id] = module_obj

    def store_port(
        self, exact_port_id: str, port_obj: GenericL23Port, port_serdes_num: int
    ) -> None:
        self._pt_state.ports[exact_port_id] = port_obj
        self._pt_state.port_serdes_num[exact_port_id] = port_serdes_num

    def retrieve_module(self, exact_module_id: str = "current") -> GenericL23Module:
        if self.retrieve_tester() is None:
            raise NotConnectedError()
        if exact_module_id == "current":
            if not self._md_state.module_str:
                raise NoWorkingPort()
            exact_module_id = self._md_state.module_str
        if exact_module_id in self._md_state.modules:
            return self._md_state.modules[exact_module_id]
        raise NotInStoreError(exact_module_id)

    def retrieve_port(self, exact_port_id: str = "current") -> GenericL23Port:
        if self.retrieve_tester() is None:
            raise NotConnectedError()
        if exact_port_id == "current":
            if not self._pt_state.port_str:
                raise NoWorkingPort()
            exact_port_id = self._pt_state.port_str
        if exact_port_id in self._pt_state.ports:
            return self._pt_state.ports[exact_port_id]
        raise NotInStoreError(exact_port_id)

    def validate_current_serdes(self, serdes: int) -> None:
        current_port_id = self._pt_state.port_str
        if current_port_id not in self._pt_state.port_serdes_num:
            raise NotInStoreError(current_port_id)
        if serdes not in range(self._pt_state.port_serdes_num[self._pt_state.port_str]):
            raise NotCorrectSerdesError(current_port_id, serdes)

    def remove_module(self, exact_module_id: str) -> None:
        if exact_module_id in self._md_state.modules:
            del self._md_state.modules[exact_module_id]

    def remove_port(self, exact_port_id: str) -> None:
        if exact_port_id in self._pt_state.ports:
            del self._pt_state.ports[exact_port_id]

    def remove_ports(self) -> None:
        self._pt_state = PortState()

    def set_error(self, error_str: str) -> None:
        self._error.set_val(error_str)

    clear_error = partialmethod(set_error, "")

    def get_error(self) -> str:
        return self._error.err_str

    def obtain_physical_modules(
        self, id_str: str = "*", update=True
    ) -> dict[str, GenericL23Module]:
        tester = self.retrieve_tester()
        if tester is None:
            raise NotConnectedError()
        m_dics = {}
        if id_str == "*":
            for m in mgmt_utils.get_modules(tester):
                m_dics[str(m.module_id)] = m
        else:
            module_id = id_str.split("/")[0]
            m = mgmt_utils.get_module(tester, int(module_id))
            m_dics[str(m.module_id)] = m

        if update:
            self._md_state.modules.update(m_dics)
        return m_dics

    def obtain_physical_ports(
        self, id_str: str = "*", update: bool = True
    ) -> dict[str, GenericL23Port]:
        tester = self.retrieve_tester()
        if tester is None:
            raise NotConnectedError()

        p_dics = {}
        if id_str == "*":
            m_id = p_id = -1
            for i in mgmt_utils.get_all_ports(tester):
                p_dics[f"{i.kind.module_id}/{i.kind.port_id}"] = i
        else:
            splitted = id_str.split("/")
            if len(splitted) == 1:
                m_id = splitted[0]
                p_id = -1
                try:
                    m_id = int(m_id)
                    for i in mgmt_utils.get_ports(tester, m_id):
                        p_dics[f"{i.kind.module_id}/{i.kind.port_id}"] = i
                except ValueError:
                    raise NoSuchIDError(id_str)
            elif len(splitted) == 2:
                m_id, p_id = splitted
                try:
                    m_id = int(m_id)
                    p_id = int(p_id)
                    i = mgmt_utils.get_port(tester, m_id, p_id)
                    p_dics[f"{i.kind.module_id}/{i.kind.port_id}"] = i
                except ValueError:
                    raise NoSuchIDError(id_str)
            else:
                raise NoSuchIDError(id_str)
        if update:
            self._pt_state.ports.update(p_dics)
        return p_dics



import setuptools


def main():
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()

    setuptools.setup(
        name="xoa-utils",
        entry_points={
            "console_scripts": [
                "xoa-utils = xoa_utils.entry:main",
            ]
        },
        description=(
            "Xena OpenAutomation Utilities provides a shell-like command-line interface for users to do explorative"
            " tests interactively, such as ANLT test."
        ),
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="Ron Ding, Leonard Yu",
        author_email="rdi@xenanetworks.com, hyu@xenanetworks.com",
        maintainer="Xena Networks",
        maintainer_email="support@xenanetworks.com",
        url="https://github.com/xenanetworks/open-automation-utilities",
        packages=setuptools.find_packages(),
        license="Apache 2.0",
        install_requires=[
            "typing_extensions>=4.4.0",
            "cffi>=1.15.1",
            "cryptography>=39.0.0",
            "pycparser>=2.21",
            "colorama>=0.4.6",
            "idna>=3.4",
            "asyncssh>=2.13.0",
            "asyncclick>=8.1.3.4",
            "anyio>=3.6.2",
            "loguru>=0.6.0",
            "pdoc>=12.3.1",
            "pytest>=7.2.1",
            "asyncclick>=8.1.3.4",
            "psutil>=5.9.4",
            "xoa-driver>=1.3",
        ],
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
        ],
        python_requires=">=3.8",
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

# import configparser
import os
import typing as t
import asyncclick as ac
from xoa_driver import enums

if t.TYPE_CHECKING:
    from xoa_driver.ports import GenericL23Port
    from xoa_utils.cmds.cmd_context import CmdContext


class ReadConfig:
    def __init__(
        self,
        ssh_port: str = "22622",
        ssh_key_path: str = os.path.join(os.path.expanduser("~"), ".ssh", "id_rsa"),
        hub_host: str = "localhost",
        hub_port: str = "10000",
        hub_enable: str = "false",
        hub_pid_path: str = os.path.join(
            os.path.expanduser("~"), "XenaNetworks", "XOA-UTILITIES", "hub.pid"
        ),
    ) -> None:
        try:
            ssh_port_int = int(ssh_port)
        except Exception:
            ssh_port_int = 22622
        try:
            hub_port_int = int(hub_port)
        except Exception:
            hub_port_int = 10000
        self.conn_port = ssh_port_int
        self.conn_host_keys = ssh_key_path
        self.hub_enabled = str(hub_enable).lower() == "true"
        self.hub_host = hub_host
        self.hub_port = hub_port_int
        self.hub_pid_path = hub_pid_path
        self.hub_pid = self._read_hub_pid()

    def _touch(self, realpath: str) -> None:
        realpath = os.path.abspath(realpath)
        folder = os.path.dirname(realpath)
        if not os.path.isfile(realpath):
            try:
                os.remove(realpath)
            except Exception:
                pass
            os.makedirs(folder, exist_ok=True)
            with open(realpath, "w"):
                pass

    def _read_hub_pid(self) -> int:
        pid = 0
        self._touch(self.hub_pid_path)
        with open(self.hub_pid_path, "r") as f:
            content = f.read().strip()
            try:
                pid = int(content)
            except ValueError:
                pass
        return pid


def run_coroutine_as_sync(coro: t.Coroutine) -> t.Any:
    while True:
        try:
            next(coro.__await__())
        except StopIteration as done:
            return done.value


def format_error(error: ac.UsageError) -> str:
    hint = ""
    result = ""
    if (
        error.ctx is not None
        and error.ctx.command.get_help_option(error.ctx) is not None
    ):
        hint = f"Try '{error.ctx.command_path} {error.ctx.help_option_names[0]}' for help.\n\n"
    if error.ctx is not None:
        result += f"{error.ctx.get_usage()}\n{hint}"
    result += f"Error: {error.format_message()}\n"
    result = result.replace("python -m entry", "")
    return result


def _port_dic_status(current_id: str, port_dic: dict[str, GenericL23Port]) -> str:
    string = f"Port      Sync      Owner     \n"
    for name, port in port_dic.items():
        new_name = f"*{name}" if current_id == name else name
        owner = "You" if port.is_reserved_by_me() else "Others"
        sync_status = str(port.info.sync_status.name)
        string += f"{new_name:10}{sync_status:10}{owner:10}\n"

    return string


def format_tester_status(storage: "CmdContext") -> str:
    serial_number = storage.retrieve_tester_serial()
    con_info = storage.retrieve_tester_con_info()
    username = storage.retrieve_tester_username()
    result_str = f"\nTester  :      {serial_number}\nConInfo :      {con_info}\nUsername:      {username}\n\n"
    result_str += _port_dic_status(
        storage.retrieve_port_str(), storage.retrieve_ports()
    )
    return result_str


def format_ports_status(storage: "CmdContext", all: bool) -> str:
    if all:
        port_dic = storage.get_all_ports()
    else:
        port_dic = storage.retrieve_ports()
    result_str = _port_dic_status(storage.retrieve_port_str(), port_dic)
    return result_str


def format_port_status(status: dict, storage: "CmdContext") -> str:
    return f"""
[ACTUAL CONFIG]
    Link recovery         : {status['link_recovery']}
    Serdes count          : {status['serdes_count']}

    Auto-negotiation      : {status['autoneg_enabled']} ({'allow' if status['autoneg_allow_loopback'] else 'not allow'} loopback)
    Link training         : {'on' if status['link_training_mode'] != "disabled" else 'off'} ({'interactive' if status['link_training_mode'] == "interactive" else 'auto'}) (preset0: {'standard tap' if status['link_training_preset0'] == 'nrz_no_preset' else 'existing tap'} values) (timeout: {status['link_training_timeout']})
    

[SHADOW CONFIG]
    Auto-negotiation      : {'on' if storage.retrieve_an_enable() else 'off'} ({'allow' if storage.retrieve_an_loopback() else 'not allow'} loopback)
    Link training         : {'on' if storage.retrieve_lt_enable() else 'off'} ({'interactive' if storage.retrieve_lt_interactive() else 'auto'}) (preset0: {'standard tap' if storage.retrieve_lt_preset0() == enums.NRZPreset.NRZ_NO_PRESET else 'existing tap'} values)
"""


def format_an_status(dic: dict) -> str:
    return f"""
[AN STATUS]
    Mode                  : {'enabled' if dic['is_enabled'] else 'disabled'}
    Loopback              : {dic['loopback']}

    Duration              : {dic['duration']:,} µs {'(N/A)' if dic['duration'] == 0 else ''}

    Successful runs       : {dic['successes']}
    Timeouts              : {dic['timeouts']}
    Loss of sync          : {dic['loss_of_sync']}
    FEC negotiation fails : {dic['fec_negotiation_fails']}
    HCD negotiation fails : {dic['hcd_negotiation_fails']}
                                RX    TX
    Link codewords        : {dic['link_codewords']['rx']:6}{dic['link_codewords']['tx']:6}
    Next-page messages    : {dic['next_page_messages']['rx']:6}{dic['next_page_messages']['tx']:6}
    Unformatted pages     : {dic['unformatted_pages']['rx']:6}{dic['unformatted_pages']['tx']:6}
    """


def format_lt_config(storage: CmdContext) -> str:
    return f"""
LT configuration to be on port {storage.retrieve_port_str()}
[SHADOW CONFIG]
    Link training         : {'on' if storage.retrieve_lt_enable() else 'off'} ({'interactive' if storage.retrieve_lt_interactive() else 'auto'}) (preset0: {'standard tap' if storage.retrieve_lt_preset0() == enums.NRZPreset.NRZ_NO_PRESET else 'existing tap'} values)
"""


def format_lt_im(status: dict, storage: CmdContext, serdes: int) -> str:
    ims = {}
    algs = {}

    for key, val in storage.retrieve_lt_initial_mod().items():
        ims[key] = enums.LinkTrainEncoding(val).name
    for key, val in storage.retrieve_lt_algorithm().items():
        algs[key] = enums.LinkTrainAlgorithm(val).name

    return f"""
[ACTUAL CONFIG]
    Link training         :
        Initial Mod.      : {status['initial_mods']}

Initial modulation to be {storage.retrieve_lt_initial_mod_serdes(serdes).name} on Serdes {serdes}
[SHADOW CONFIG]
    Link training         : {'on' if storage.retrieve_lt_enable() else 'off'} ({'interactive' if storage.retrieve_lt_interactive() else 'auto'}) (preset0: {'standard tap' if storage.retrieve_lt_preset0() == enums.NRZPreset.NRZ_NO_PRESET else 'existing tap'} values)
        Initial Mod.      : {ims}
    """


def format_lt_algorithm(status: dict, storage: CmdContext, serdes: int) -> str:
    ims = {}
    algs = {}

    for key, val in storage.retrieve_lt_initial_mod().items():
        ims[key] = enums.LinkTrainEncoding(val).name
    for key, val in storage.retrieve_lt_algorithm().items():
        algs[key] = enums.LinkTrainAlgorithm(val).name

    return f"""
[ACTUAL CONFIG]
    Link training         :
        Algorithm         : {status['algorithms']}

LT algorithm to be {storage.retrieve_lt_algorithm_serdes(serdes).name} on Serdes {serdes}
[SHADOW CONFIG]
    Link training         : {'on' if storage.retrieve_lt_enable() else 'off'} ({'interactive' if storage.retrieve_lt_interactive() else 'auto'}) (preset0: {'standard tap' if storage.retrieve_lt_preset0() == enums.NRZPreset.NRZ_NO_PRESET else 'existing tap'} values)
        Algorithm         : {algs}
    """


def format_an_config(storage: CmdContext) -> str:

    return f"""
AN configuration to be on port {storage.retrieve_port_str()}
[SHADOW CONFIG]
    Auto-negotiation      : {'on' if storage.retrieve_an_enable() else 'off'} ({'allow' if storage.retrieve_an_loopback() else 'not allow'} loopback)
    Link training         : {'on' if storage.retrieve_lt_enable() else 'off'} ({'interactive' if storage.retrieve_lt_interactive() else 'auto'}) (preset0: {'standard tap' if storage.retrieve_lt_preset0() == enums.NRZPreset.NRZ_NO_PRESET else 'existing tap'} values)
"""


def format_recovery(storage: CmdContext, on: bool) -> str:
    enable = "on" if on else "off"
    return f"Port {storage.retrieve_port_str()} link recovery: {enable}\n"


def format_lt_inc_dec(
    storage: CmdContext, serdes: int, emphasis: str, increase: bool, response: str
) -> str:
    change = {
        "pre3": "c(-3)",
        "pre2": "c(-2)",
        "pre": "c(-1)",
        "main": "c(0)",
        "post": "c(1)",
    }[emphasis]
    action = "increase" if increase else "decrease"
    return f"Port {storage.retrieve_port_str()}: {action} {change} by 1 on Serdes {serdes} ({response})\n"


def format_lt_encoding(
    storage: CmdContext, serdes: int, encoding: str, response: str
) -> str:
    e = enums.LinkTrainEncoding[
        {"pam4pre": "PAM4_WITH_PRECODING"}.get(encoding, encoding).upper()
    ]
    return f"Port {storage.retrieve_port_str()}: use {e.name} on Serdes {serdes} ({response})\n"


def format_lt_preset(
    storage: CmdContext, serdes: int, preset: int, response: str
) -> str:
    return f"Port {storage.retrieve_port_str()}: use preset {preset} on Serdes {serdes} ({response})\n"


def format_lt_trained(storage: CmdContext, serdes: int, response: str) -> str:
    return f"Port {storage.retrieve_port_str()} requests: Serdes {serdes} is trained ({response})\n"


def format_txtap_get(serdes: int, dic: dict) -> str:
    return f"""
Local Coefficient Serdes({serdes}) :           c(-3)       c(-2)       c(-1)        c(0)        c(1)
    Current level           :              {dic['c(-3)']}           {dic['c(-2)']}           {dic['c(-1)']}           {dic['c(0)']}           {dic['c(1)']}
"""


def format_txtap_set(
    serdes: int, pre3: int, pre2: int, pre: int, main: int, post: int
) -> str:
    return format_txtap_get(
        serdes, {"c(-3)": pre3, "c(-2)": pre2, "c(-1)": pre, "c(0)": main, "c(1)": post}
    )


def format_lt_status(dic: dict) -> str:
    return f"""
[LT STATUS]
    Mode              : {'enabled' if dic['is_enabled'] else 'disabled'}
    Status            : {'trained' if dic['is_trained'] else 'not trained'}
    Failure           : {dic['failure']}

    Initial mod.      : {dic['init_modulation']}
    Preset0 (oos)     : {"standard tap" if dic['preset0'] else "existing tap"} values

    Total bits        : {dic['total_bits']:,}
    Total err. bits   : {dic['total_errored_bits']:,}
    BER               : {dic['ber']}

    Duration          : {dic['duration']:,} µs {'(N/A)' if dic['duration'] == 0 else ''}

    Lock lost         : {dic['lock_lost']}
    Frame lock        : {dic['frame_lock']}
    Remote frame lock : {dic['remote_frame_lock']}

    Frame errors      : {dic['frame_errors']:,}
    Overrun errors    : {dic['overrun_errors']:,}

    Last IC received  : {dic['last_ic_received']}
    Last IC sent      : {dic['last_ic_sent']}

    TX Coefficient              :          c(-3)       c(-2)       c(-1)        c(0)        c(1)
        Current level           :{dic['c(-3)']['current_level']:15}{dic['c(-2)']['current_level']:12}{dic['c(-1)']['current_level']:12}{dic['c(0)']['current_level']:12}{dic['c(1)']['current_level']:12}
                                :         RX  TX      RX  TX      RX  TX      RX  TX      RX  TX
        + req                   :{dic['c(-3)']['+req']['rx']:11}{dic['c(-3)']['+req']['tx']:4}{dic['c(-2)']['+req']['rx']:8}{dic['c(-2)']['+req']['tx']:4}{dic['c(-1)']['+req']['rx']:8}{dic['c(-1)']['+req']['tx']:4}{dic['c(0)']['+req']['rx']:8}{dic['c(0)']['+req']['tx']:4}{dic['c(1)']['+req']['rx']:8}{dic['c(1)']['+req']['tx']:4}
        - req                   :{dic['c(-3)']['-req']['rx']:11}{dic['c(-3)']['-req']['tx']:4}{dic['c(-2)']['-req']['rx']:8}{dic['c(-2)']['-req']['tx']:4}{dic['c(-1)']['-req']['rx']:8}{dic['c(-1)']['-req']['tx']:4}{dic['c(0)']['-req']['rx']:8}{dic['c(0)']['-req']['tx']:4}{dic['c(1)']['-req']['rx']:8}{dic['c(1)']['-req']['tx']:4}
        coeff/eq limit reached  :{dic['c(-3)']['coeff_and_eq_limit_reached']['rx']:11}{dic['c(-3)']['coeff_and_eq_limit_reached']['tx']:4}{dic['c(-2)']['coeff_and_eq_limit_reached']['rx']:8}{dic['c(-2)']['coeff_and_eq_limit_reached']['tx']:4}{dic['c(-1)']['coeff_and_eq_limit_reached']['rx']:8}{dic['c(-1)']['coeff_and_eq_limit_reached']['tx']:4}{dic['c(0)']['coeff_and_eq_limit_reached']['rx']:8}{dic['c(0)']['coeff_and_eq_limit_reached']['tx']:4}{dic['c(1)']['coeff_and_eq_limit_reached']['rx']:8}{dic['c(1)']['coeff_and_eq_limit_reached']['tx']:4}
        eq limit reached        :{dic['c(-3)']['eq_limit_reached']['rx']:11}{dic['c(-3)']['eq_limit_reached']['tx']:4}{dic['c(-2)']['eq_limit_reached']['rx']:8}{dic['c(-2)']['eq_limit_reached']['tx']:4}{dic['c(-1)']['eq_limit_reached']['rx']:8}{dic['c(-1)']['eq_limit_reached']['tx']:4}{dic['c(0)']['eq_limit_reached']['rx']:8}{dic['c(0)']['eq_limit_reached']['tx']:4}{dic['c(1)']['eq_limit_reached']['rx']:8}{dic['c(1)']['eq_limit_reached']['tx']:4}
        coeff not supported     :{dic['c(-3)']['coeff_not_supported']['rx']:11}{dic['c(-3)']['coeff_not_supported']['tx']:4}{dic['c(-2)']['coeff_not_supported']['rx']:8}{dic['c(-2)']['coeff_not_supported']['tx']:4}{dic['c(-1)']['coeff_not_supported']['rx']:8}{dic['c(-1)']['coeff_not_supported']['tx']:4}{dic['c(0)']['coeff_not_supported']['rx']:8}{dic['c(0)']['coeff_not_supported']['tx']:4}{dic['c(1)']['coeff_not_supported']['rx']:8}{dic['c(1)']['coeff_not_supported']['tx']:4}
        coeff at limit          :{dic['c(-3)']['coeff_at_limit']['rx']:11}{dic['c(-3)']['coeff_at_limit']['tx']:4}{dic['c(-2)']['coeff_at_limit']['rx']:8}{dic['c(-2)']['coeff_at_limit']['tx']:4}{dic['c(-1)']['coeff_at_limit']['rx']:8}{dic['c(-1)']['coeff_at_limit']['tx']:4}{dic['c(0)']['coeff_at_limit']['rx']:8}{dic['c(0)']['coeff_at_limit']['tx']:4}{dic['c(1)']['coeff_at_limit']['rx']:8}{dic['c(1)']['coeff_at_limit']['tx']:4}
"""

def format_debug_init(dic: dict) -> str:
    return f"""
    base:        0x{dic["base"]:0>8X}
    rx_gtm_base: 0x{dic["rx_gtm_base"]:0>8X}
    rx_serdes:   {dic["rx_serdes"]}
    tx_gtm_base: 0x{dic["tx_gtm_base"]:0>8X}
    tx_serdes:   {dic["tx_serdes"]}
"""

def format_strict(storage: CmdContext, on: bool) -> str:
    enable = "on" if on else "off"
    return f"Port {storage.retrieve_port_str()} ANLT strict mode: {enable}\n"

def format_log_control(
        storage: CmdContext, 
        debug: bool, 
        an_trace: bool, 
        lt_trace: bool, 
        alg_trace: bool,
        fsm_port: bool,
        fsm_an: bool,
        fsm_an_stimuli: bool,
        fsm_lt: bool,
        fsm_lt_coeff: bool,
        fsm_lt_stimuli: bool,
        fsm_lt_alg0: bool,
        fsm_lt_algn1: bool
        ) -> str:
    return f"""
Port {storage.retrieve_port_str()} log control:
    Type debug:             {'on' if debug else 'off'}
    Type AN trace:          {'on' if an_trace else 'off'}
    Type LT trace:          {'on' if lt_trace else 'off'}
    Type ALG trace:         {'on' if alg_trace else 'off'}
    Type FSM port:          {'on' if fsm_port else 'off'}
    Type FSM AN:            {'on' if fsm_an else 'off'}
    Type FSM AN Stimuli:    {'on' if fsm_an_stimuli else 'off'}
    Type FSM LT:            {'on' if fsm_lt else 'off'}
    Type FSM LT Coeff:      {'on' if fsm_lt_coeff else 'off'}
    Type FSM LT Stimuli:    {'on' if fsm_lt_stimuli else 'off'}
    Type FSM LT ALG  0:     {'on' if fsm_lt_alg0 else 'off'}
    Type FSM LT ALG -1:     {'on' if fsm_lt_algn1 else 'off'}
"""